import typer

from nef_pipelines import nef_app

help_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        help_app, name="help", help="- help on the nef pipelines tools and their usage"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.help.about  # noqa: F401
    import nef_pipelines.tools.help.commands  # noqa: F401



import contextlib
import sys
from pathlib import Path
from textwrap import dedent
from typing import List, Optional

import typer
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import chains_from_frames, count_residues
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    strings_to_table_terminal_sensitive,
)
from nef_pipelines.tools.frames import frames_app

UNDERSCORE = "_"

parser = None


# noinspection PyUnusedLocal
@frames_app.command()
def list(
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="read NEF data from a file instead of stdin",
    ),
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-t",
        "--selector-type",
        help=f"force how to select frames can be one of {', '.join(SelectionType.__members__)}",
    ),
    number: bool = typer.Option(False, "-n", "--number", help="number entries"),
    verbose: int = typer.Option(
        False,
        "-v",
        "--verbose",
        count=True,
        help="print verbose information more verbose options give more information",
    ),
    write_error: bool = typer.Option(
        False, "--write-error", help="write output to stderr"
    ),
    one_per_line: bool = typer.Option(
        False, "-1", "--one-per-line", help="write frame names one per line"
    ),
    filters: Optional[List[str]] = typer.Argument(
        None, help="filters string for entry names and categories to list"
    ),
):
    """- list the frames in the current input"""

    entry = None
    if len(filters) > 0:
        entry = _if_is_nef_file_load_as_entry(filters[0])
        if entry is not None:
            if input != STDIN:
                msg = f"""\
                   two nef file paths supplied...
                       path 1: {input} [from --in]
                       path 2: {filters[0]} [from args]"""
                msg = dedent(msg)
                exit_error(msg)
            else:
                input = filters[0]
                filters = filters[1:]
    if not filters:
        filters = [
            "*",
        ]

    args = get_args()

    if entry is None:

        try:
            entry = read_entry_from_file_or_stdin_or_exit_error(input)
        except Exception as e:
            exit_error(f"failed to read nef file {args.pipe} because", e)

    if entry is None:
        if args.pipe is not None:
            exit_error(f"couldn't read a nef stream from the file: {args.pipe}")
        elif len(filters) > 0:
            exit_error(f"couldn't read a nef stream from either {filters[0]} or stdin")
        else:
            exit_error("couldn't read a nef stream from stdin")

    lines = str(entry)
    output_file = sys.stderr if write_error else sys.stdout
    with contextlib.redirect_stdout(output_file):

        if verbose > 0:

            import hashlib

            print(f"entry {entry.entry_id}")
            print()

            md5 = hashlib.md5(lines.encode("ascii")).hexdigest()
            num_lines = len(lines.split("\n"))
            print(f"lines: {num_lines} frames: {len(entry)} checksum: {md5} [md5]")
            print()

        frames = select_frames(entry, filters, selector_type)

        if verbose == 0:
            frame_names = [frame.name for frame in frames]
            if number:
                frame_names = [
                    f"{i}. {frame_name}"
                    for i, frame_name in enumerate(frame_names, start=1)
                ]

            if one_per_line:
                frame_list = [
                    [
                        frame_name,
                    ]
                    for frame_name in frame_names
                ]
            else:
                frame_list = strings_to_table_terminal_sensitive(frame_names)
            print(tabulate(frame_list, tablefmt="plain"))

        else:

            for i, frame in enumerate(frames, start=1):

                filters = [f"*{filter}*" for filter in filters]

                if verbose > 0:
                    frame_names = [
                        f"{i}. {frame.name}"
                        for i, frame_name in enumerate(frames, start=1)
                    ]

                    frame_list = strings_to_table_terminal_sensitive(frame_names)
                    print(f"{i}. {frame.name}")
                    print(f"    category: {frame.category}")
                    if len(frame.name) != len(frame.category):
                        print(
                            f"    name: {frame.name[len(frame.category):].lstrip(UNDERSCORE)}"
                        )

                    loop_lengths = []
                    for loop in frame.loop_dict:
                        loop_lengths.append(str(len(loop)))

                    loops = ""
                    if len(loops) == 1:
                        loops = " [length: 1]"
                    else:
                        loops = f' [lengths: {", ".join(loop_lengths)}]'

                    print(f"    loops: {len(frame.loop_dict)}{loops}")

                    loop_names = [name.lstrip("_") for name in frame.loop_dict.keys()]
                    comma = ", "
                    print(f"    loop names: {comma.join(loop_names)}")

                    frame_standard = frame.name[: len("nef_")]
                    is_standard_frame = frame_standard == "nef_"
                    print(f"    is nef frame: {is_standard_frame}")

                    # ccpn_compound_name
                    if verbose == 2 and frame.category == "nef_molecular_system":
                        chains = chains_from_frames(frame)

                        print(f'    chains: {len(chains)} [{", ".join(chains)}]')

                        residue_counts = {}
                        for chain in chains:
                            residue_counts[chain] = count_residues(frame, chain)

                        residue_count_per_chain = {}
                        for chain in chains:
                            residue_count_per_chain[chain] = sum(
                                residue_counts[chain].values()
                            )

                        output = [
                            f"{chain} {num_residues}"
                            for chain, num_residues in residue_count_per_chain.items()
                        ]
                        print(f'    residues: {", ".join(output)}')

                        for chain in chains:
                            counts_and_percentages = []

                            for residue, count in residue_counts[chain].items():
                                percentage = (
                                    f"{count/ residue_count_per_chain[chain]*100:5.2f}"
                                )
                                counts_and_percentages.append(
                                    f"{residue}: {count} [{percentage}%]"
                                )

                            pre_string = f"              {chain}. "
                            pre_string_width = len(pre_string)

                            tabulation = strings_to_table_terminal_sensitive(
                                counts_and_percentages, pre_string_width
                            )
                            table = tabulate(tabulation, tablefmt="plain")

                            print(_indent_with_prestring(table, pre_string))

                print()

        print()


def _indent_with_prestring(text_block, pre_string):
    raw_result = []
    empty_prestring = " " * len(pre_string)
    for i, string in enumerate(text_block.split("\n")):
        if i == 0:
            raw_result.append(f"{pre_string}{string}")
        else:
            raw_result.append(f"{empty_prestring}{string}")

    return "\n".join(raw_result)


def _if_is_nef_file_load_as_entry(file_path):
    entry = None
    try:
        entry = Entry.from_file(file_path)
    except Exception:
        pass

    return entry



import string
from dataclasses import dataclass
from enum import auto
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry
from strenum import LowercaseStrEnum
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import (
    NEF_MOLECULAR_SYSTEM,
    SELECTORS_LOWER,
    SelectionType,
    loop_row_dict_iter,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_ends,
    get_chain_starts,
    sequences_from_frames,
)
from nef_pipelines.lib.util import (
    chunks,
    exit_error,
    is_int,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
    strip_characters_right,
)
from nef_pipelines.tools.loops import loops_app


class TrimAction(LowercaseStrEnum):
    DELETE = auto()
    DEASSIGN = auto()


@dataclass
class ChainBound:
    chain: str
    start: int
    end: int


# noinspection PyUnusedLocal
@loops_app.command()
def trim(
    chain_bounds: List[str] = Argument(
        None,
        metavar="<CHAIN-CODE> <START> <END> | <FRAME_SELECTOR>",
        help="""chain-codes and inclusive chain starts and ends to trim the chain to
                or a list of reference frames to find chain bounds from
                [default: use the molecular system.]""",
        show_default=False,
    ),
    reference_selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-t",
        "--reference-selector-type",
        help=f"control how to select reference frames to find chain bounds, can be one of: {SELECTORS_LOWER}. ",
    ),
    input: Path = typer.Option(
        Path("-"),
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="where to read NEF data from either a file or stdin '-'",
    ),
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-s",
        "--selector",
        help=f"control how to select frames to trim, can be one of: {SELECTORS_LOWER}. "
        "Any will match on names first and then if there is no match attempt to match on category",
    ),
    # invert: bool = Option(False,  "--invert", help="invert the selection"),
    # action: TrimAction = Option(
    #     TrimAction.DELETE,
    #     "--action",
    #     metavar="<ACTION>",
    #     help="the action to take on the selected chains [default: renumber]",
    # ),
    frame_selectors: List[str] = Option(
        None,
        "-f",
        "--frames",
        metavar="<FRAME-SELECTOR>",
        help="Limit changes to a particular frame by name or category [the selector], note: wildcards are "
        "allowed. Frames are selected by name and then category if the name doesn't match"
        "note: the option -t /--selector-type allows you to choose which selection type to use. "
        "[default: all frames]",
        show_default=False,
    ),
):
    """- trim the lengths of assigned chains in a nef file [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    chain_bounds = parse_comma_separated_options(chain_bounds)

    if not chain_bounds:
        chain_bounds = [
            NEF_MOLECULAR_SYSTEM,
        ]

    chains = chains_from_frames(entry.frame_list)

    reference_frames = _find_reference_frames(
        entry, chain_bounds, reference_selector_type
    )

    if reference_frames:
        chain_bounds = _chain_bounds_from_frames(reference_frames)
    else:
        chain_bounds = _parse_chain_bounds_exit_if_bad(chain_bounds, chains)

    entry = pipe(entry, frame_selectors, selector_type, chain_bounds)

    print(entry)


def pipe(
    entry: Entry,
    frame_selectors: List[str],
    frame_selector_type: SelectionType,
    chain_bounds: List[ChainBound],
) -> Entry:

    frames = select_frames(entry, frame_selectors, frame_selector_type)

    target_chains = chain_bounds.keys()

    _exit_if_selected_chain_not_in_frames("chain to trim", frames, input, target_chains)

    _trim_chains_in_frames(frames, chain_bounds)

    return entry


def _find_reference_frames(entry, chain_bounds, reference_selector_type):
    return select_frames(entry, chain_bounds, reference_selector_type)


def _chain_bounds_from_frames(reference_frames):
    sequences = sequences_from_frames(reference_frames)
    chain_starts = get_chain_starts(sequences)
    chain_ends = get_chain_ends(sequences)

    result = {}
    for chain in chain_starts:
        result[chain] = [
            ChainBound(chain, chain_starts[chain], chain_ends[chain]),
        ]

    return result


def _trim_chains_in_frames(frames, chain_bounds):

    for frame in frames:
        rows_to_remove = set()

        for loop in frame:
            for i, row in enumerate(loop_row_dict_iter(loop)):
                for name, value in row.items():
                    if name.startswith("chain_code"):
                        _, index = strip_characters_right(name, string.digits + "_")
                        chain_code = value
                        if chain_code not in chain_bounds:
                            continue

                        sequence_code_tag = f"sequence_code{index}"
                        sequence_code = row[sequence_code_tag]
                        if is_int(sequence_code):
                            sequence_code = int(sequence_code)
                        else:
                            continue

                        for chain_bound in chain_bounds[chain_code]:
                            if (
                                sequence_code < chain_bound.start
                                or sequence_code > chain_bound.end
                            ):
                                rows_to_remove.add(i)

        rows_to_keep = []
        for i, row in enumerate(loop_row_dict_iter(loop)):
            if i not in rows_to_remove:
                rows_to_keep.append(dict(row))

        loop.clear_data()

        loop.add_data(rows_to_keep)

        if "index" in loop.tags:
            loop.renumber_rows("index")


def _exit_error_if_no_chain_bounds(chain_bounds):
    if not chain_bounds:
        msg = """
            you didn't provide any chain bounds to trim the chains
            """
        exit_error(msg)


def _exit_if_selected_chain_not_in_frames(chain_type, frames, input, chains):
    frame_chains = set()
    for frame in frames:
        frame_chains.update(chains_from_frames(frame))
    frame_chain_names = strings_to_tabulated_terminal_sensitive(frame_chains)

    for chain in chains:

        selected_frames = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in frames]
        )
        if chain not in frame_chains:
            msg = """
                using the input {input}
                the {chain_type} '{chain}' is not in the frames selected for renumbering

                the selected frames are:

                {selected_frames}

                which have the chains

                {frame_chain_names}
                """
            msg = msg.format(
                chain_type=chain_type,
                input=input,
                chain=chain,
                selected_frames=selected_frames,
                frame_chain_names=frame_chain_names,
            )
            exit_error(msg)


def _parse_chain_bounds_exit_if_bad(chain_bounds, chains):

    if len(chain_bounds) % 3:
        msg = """
            the chain bounds must be a triplet of values
            each chain bound being of the form <chain-code> <start> <end>
            you gave {chain_bounds}
            """
        msg = msg.format(chain_bounds=chain_bounds)
        exit_error(msg)

    result = {}
    for chain, start, end in chunks(chain_bounds, 3):
        if chain not in chains:
            msg = f"""
                the chain {chain} is not in the chains in the read frames
                the chains in the stream are {', '.join(chains)}
                """
            msg = msg.format(chain=chain, chains=chains)
            exit_error(msg)

        if not is_int(start):
            msg = f"""
                the start value {start} for chain {chain} is not an integer
                """
            exit_error(msg)

        if not is_int(end):
            msg = f"""
                the end value {end} for chain {chain} is not an integer
                """
            exit_error(msg)

        result.setdefault(chain, []).append(ChainBound(chain, int(start), int(end)))

        if end <= start:
            msg = f"""
                the end value {end} for chain {chain} is not greater than the start value {start}
                """
            exit_error(msg)

    return result



import sys
from math import sqrt
from textwrap import dedent
from typing import Dict, List

import typer
from pynmrstar import Loop, Saveframe
from pynmrstar.entry import Entry

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    create_nef_save_frame,
    loop_row_dict_iter,
    read_entry_from_stdin_or_exit,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.structures import AtomLabel, NewPeak, SequenceResidue, ShiftData
from nef_pipelines.lib.util import exit_error, flatten
from nef_pipelines.tools.peaks import peaks_app
from nef_pipelines.transcoders.nmrview.importers.shifts import add_frames_to_entry

# CF M P Williamson PROGRESS IN NUCLEAR MAGNETIC RESONANCE SPECTROSCOPY 2014
# Using chemical shift perturbation to characterise ligand binding” [Prog. Nucl. Magn. Reson. Spectrosc. 73C (2013) 1–16
DEFAULT_ISOTOPE_WEIGHTS = weights = {"1H": 1.0, "15N": 7.0, "13C": 3.333333}


@peaks_app.command()
def match(
    # ignore_amide_shifts: bool = typer.Option(False, help='ignore amides when matching'),
    # amide_weight: float = typer.Option(1.0, help='weight for the amide shifts'),
    # amide_search_region: Tuple[float,float] = typer.Option((0.2, 0.2*7), help='region to search for amide shifts in
    # the order H N'),
    assign: bool = typer.Option(
        False,
        "--assign",
        help="assign the second peak list using closest matches from the first",
    ),
    names: List[str] = typer.Argument(
        ..., help="pairs of shift frame names for the chemical shifts to compare"
    ),
):
    """- match one list of peaks against another [alpha]"""

    print(
        "*** WARNING *** this command [peaks match] is only lightly tested use at your own risk!",
        file=sys.stderr,
    )

    if len(names) != 2:
        exit_error("i currently need pairs of peaks")

    entry = read_entry_from_stdin_or_exit()

    name_1, name_2 = names

    frame_1 = _select_single_frame_or_exit(entry, name_1)
    frames_2 = _select_single_frame_or_exit(entry, name_2)

    # amide_search_region = {'H': amide_search_region[0], 'N': amide_search_region[1]}

    entry = pipe(entry, frame_1, frames_2, assign)

    print(entry)


def _nef_frames_to_peak_shifts(frame: Saveframe) -> Dict[str, Dict[str, float]]:
    loop = frame.get_loop("nef_peak")

    indices = [
        int(tag.split("_")[-1]) for tag in loop.tags if tag.startswith("atom_name")
    ]

    peak_atom_shifts = {}
    for row in loop_row_dict_iter(loop):
        key = row["peak_id"]

        atom_shifts = {}
        peak_atom_shifts[key] = atom_shifts
        for index in indices:
            shift = row[f"position_{index}"]

            atom_shifts[index - 1] = shift

    return peak_atom_shifts


def _find_best_match(
    shifts: Dict[str, float],
    match_shifts: Dict[str, Dict[str, float]],
    weights=Dict[str, float],
):

    weighted_values_1 = _get_weighted_terms(shifts, weights)

    result = {}

    for peak_id, match_shifts in match_shifts.items():
        weighted_values_2 = _get_weighted_terms(match_shifts, weights)

        terms = []

        for dim in weighted_values_1:
            terms.append((weighted_values_1[dim] - weighted_values_2[dim]) ** 2)
        distance = sqrt(sum(terms))

        result.setdefault(distance, []).append(peak_id)

    min_distance = min(result.keys())
    return result[min_distance], min_distance


def _get_weighted_terms(shifts, weights):

    return {dim: shift / weights[dim] for dim, shift in shifts.items()}


def _nef_frames_to_peak_by_id(frame):
    loop = frame.get_loop("nef_peak")

    indices = [
        int(tag.split("_")[-1]) for tag in loop.tags if tag.startswith("atom_name")
    ]

    peak_by_id = {}
    for row in loop_row_dict_iter(loop):
        peak_id = row["peak_id"]

        shifts = []
        for index in indices:
            chain_code = row[f"chain_code_{index}"]
            sequence_code = row[f"sequence_code_{index}"]
            residue_name = row[f"residue_name_{index}"]
            atom_name = row[f"atom_name_{index}"]
            position = row[f"position_{index}"]

            residue = SequenceResidue(chain_code, sequence_code, residue_name)
            atom = AtomLabel(residue, atom_name)

            shifts.append(ShiftData(atom, position))

        peak_by_id[peak_id] = NewPeak(shifts, peak_id)

    return peak_by_id


def _get_assignment_tuple(peak: NewPeak):
    result = []
    for shift in peak.shifts:
        residue = shift.atom.residue
        result.extend(
            (
                residue.chain_code,
                residue.sequence_code,
                residue.residue_name,
                shift.atom.atom_name,
            )
        )

    return tuple(result)


def _get_assignment_string(peak: NewPeak):
    result = []
    for shift in peak.shifts:
        residue = shift.atom.residue
        result.append(
            f"#{residue.chain_code}:{residue.sequence_code}[{residue.residue_name}]@{shift.atom.atom_name}"
        )

    return "-".join(result)


def _get_peak_frame_dimension_mapping(peak_frame_1, peak_frame_2):
    axis_codes_1 = _get_spectrum_axis_codes(peak_frame_1)
    axis_codes_2 = _get_spectrum_axis_codes(peak_frame_2)

    mapping = []
    for axis_code_1 in axis_codes_1:
        mapping.append(
            [i for i, code in enumerate(axis_codes_2) if code == axis_code_1]
        )
    return mapping


def _get_spectrum_axis_codes(peak_frame):
    return peak_frame.get_loop("nef_spectrum_dimension").get_tag("axis_code")


def _exit_if_mappings_bad(mapping, peak_frame_1, peak_frame_2):
    mappings_ok = [len(dim_mapping) == 1 for dim_mapping in mapping]

    if not all(mappings_ok):
        mapping_offset_by_1 = [
            [str(dim_mapping + 1) for dim_mapping in mapping_list]
            for mapping_list in mapping
        ]
        mapping_offset_by_1_joined = [
            f"{', '.join(dim_mapping)}" for dim_mapping in mapping_offset_by_1
        ]
        mapping_by_dim = [
            f"{dim} -> {dim_mapping}"
            for dim, dim_mapping in enumerate(mapping_offset_by_1_joined, start=1)
        ]
        mappings_by_dim = "\n".join(mapping_by_dim)

        msg = f"""
            The dimension mappings of dimensions in {peak_frame_1.name} -> {peak_frame_2.name}
            is not unique and singular, the mapping is:
        """
        msg = dedent(msg).strip()
        msg = f"{msg}\n\n{mappings_by_dim}"

        exit_error(msg)


def _map_peak_shifts(shifts, mapping):

    for peak_id, shifts_by_dim in shifts.items():

        mapped_shifts = {mapping[dim]: shift for dim, shift in shifts_by_dim.items()}
        shifts[peak_id] = mapped_shifts

    return shifts


def _get_dim_isotopes(peak_frame):

    result = []
    if "_nef_spectrum_dimension" in peak_frame:
        spectrum_dimension_loop = peak_frame.get_loop("nef_spectrum_dimension")
        if "axis_code" in spectrum_dimension_loop.tags:
            result = spectrum_dimension_loop.get_tag("axis_code")

    if not result:
        result = [
            None,
        ] * int(peak_frame.get_tag("num_dimensions")[0])

    return result


def _get_dim_weights(peak_frame):
    return {
        dim: DEFAULT_ISOTOPE_WEIGHTS.get(isotope, None)
        for dim, isotope in enumerate(_get_dim_isotopes(peak_frame))
    }


def _exit_if_dim_weights_not_defined(dim_weights, peak_frame_1, peak_frame_2):

    bad_dim_indices = [
        str(dim_index + 1)
        for dim_index, weight in dim_weights.items()
        if weight is None
    ]
    if bad_dim_indices:
        bad_dim_indices = ", ".join(bad_dim_indices)

        msg = f"""
            for the frames {peak_frame_1.name} and {peak_frame_2.name} appropiate isotope weigts couldn't
            be determined for the following dims {bad_dim_indices}
        """
        msg = dedent(msg).strip()
        exit_error(msg)


def _build_dim_assignments(peak):
    result = {}
    for dim_index, shift in enumerate(peak.shifts, start=1):
        result["chain_code"] = shift.atom.residue.chain_code
        result["sequence_code"] = shift.atom.residue.sequence_code
        result["residue_name"] = shift.atom.residue.residue_name

        # TODO this is hack because some internal routines use ? for unassigned not . [UNUSED]
        atom_name = shift.atom.atom_name
        atom_name = atom_name if atom_name != "?" else UNUSED
        result["atom_name"] = atom_name

    return result


def _assign_frame(result, peak_frame_2):
    for key, value in result.items():
        print(key, value)


def pipe(entry: Entry, peak_frame_1: Saveframe, peak_frame_2: Saveframe, assign=False):

    num_dimensions_1 = int(peak_frame_1.get_tag("num_dimensions")[0])

    peak_list_1 = frame_to_peaks(peak_frame_1)

    peak_list_1_by_id = {peak.id: peak for peak in peak_list_1}

    # TODO move from frames to peak dataclass instances
    _exit_if_peak_frame_dimensions_dont_match(peak_frame_1, peak_frame_2)

    mapping = _get_peak_frame_dimension_mapping(peak_frame_1, peak_frame_2)

    _exit_if_mappings_bad(mapping, peak_frame_1, peak_frame_2)

    shifts_1 = _nef_frames_to_peak_shifts(peak_frame_1)
    shifts_2 = _nef_frames_to_peak_shifts(peak_frame_2)

    mapping = flatten(mapping)

    shifts_2 = _map_peak_shifts(shifts_2, mapping)

    dim_weights = _get_dim_weights(peak_frame_1)

    _exit_if_dim_weights_not_defined(dim_weights, peak_frame_1, peak_frame_2)

    results = {}
    for peak_1_id, target_shifts in shifts_1.items():

        best_matches, distance = _find_best_match(target_shifts, shifts_2, dim_weights)

        best_matches = [best_match for best_match in best_matches]

        results[peak_1_id] = best_matches, distance

    if assign:

        target_peak_loop = peak_frame_2.get_loop("nef_peak")
        source_peak_loop = peak_frame_1.get_loop("nef_peak")

        source_values_by_index = {
            row["index"]: row for row in loop_row_dict_iter(source_peak_loop)
        }

        target_peak_new_values = {}
        for list_1_index, [target_peaks, _] in results.items():
            source_values = source_values_by_index[list_1_index]
            source_values = {
                source_name: source_value
                for source_name, source_value in source_values.items()
                if source_name.startswith("chain_code")
                or source_name.startswith("sequence_code")
                or source_name.startswith("residue_name")
                or source_name.startswith("atom_name")
            }

            for target_peak in target_peaks:
                target_peak_new_values[target_peak] = source_values

        for target_peak in loop_row_dict_iter(target_peak_loop):
            target_peak_index = target_peak["index"]
            if target_peak_index in target_peak_new_values:
                new_values = target_peak_new_values[target_peak_index]
                for column_name, new_value in new_values.items():
                    target_peak[column_name] = new_value

    else:
        result_frame = _make_result_frame(
            num_dimensions_1, peak_frame_1, peak_frame_2, peak_list_1_by_id, results
        )
        add_frames_to_entry(entry, result_frame)

    return entry


def _make_result_frame(
    num_dimensions_1, peak_frame_1, peak_frame_2, peak_list_1_by_id, result
):
    frame_category = "nefpls_chemical_shift_perturbations"
    frame_id = f"from_{peak_frame_1.name}_to_{peak_frame_2.name}"
    result_frame = create_nef_save_frame(frame_category, frame_id)
    result_frame.add_tag("frame_name_1", peak_frame_1.name)
    result_frame.add_tag("frame_name_2", peak_frame_2.name)
    matches_loop = Loop.from_scratch("nefpls_perturbations")
    atom_id_tags = []
    for comparison_id in range(1, 3):
        for dimension_id in range(1, num_dimensions_1 + 1):
            for atom_name_component in []:
                atom_id_tags.append(
                    f"{atom_name_component}_{comparison_id}_{dimension_id}"
                )
    loop_tags = (
        "index",
        "match_index",
        "peak_id_1",
        "peak_id_2",
        "distance",
        "chain_code",
        "sequence_code",
        "residue_name",
        "atom_name",
    )
    matches_loop.add_tag(loop_tags)
    result_frame.add_loop(matches_loop)
    for index, (peak_1_id, (peak_2_ids, distance)) in enumerate(
        result.items(), start=1
    ):
        assignments_peak_1 = _build_dim_assignments(peak_list_1_by_id[peak_1_id])
        for match_index, peak_2_id in enumerate(peak_2_ids, start=1):
            data = {
                "index": index,
                "match_index": match_index,
                "peak_id_1": peak_1_id,
                "peak_id_2": peak_2_id,
                "distance": distance,
                **assignments_peak_1,
            }
            matches_loop.add_data(
                [
                    data,
                ]
            )
    return result_frame


def _exit_if_peak_frame_dimensions_dont_match(peak_frame_1, peak_frame_2):
    num_dimensions_1 = int(peak_frame_1.get_tag("num_dimensions")[0])
    num_dimensions_2 = int(peak_frame_2.get_tag("num_dimensions")[0])

    if num_dimensions_1 != num_dimensions_2:
        msg = f"""
            The number of dimensions in the peak lists to match mustbe the same I got

                {peak_frame_1.name}: {num_dimensions_1}
                {peak_frame_2.name}: {num_dimensions_2}
        """
        exit_error(dedent(msg).strip())


def _shift_list_to_residue_atom_shifts(shifts_1):
    residue_shifts = {}

    for shift in shifts_1:
        chain_code = shift.atom.residue.chain_code
        sequence_code = shift.atom.residue.sequence_code
        atom_name = shift.atom.atom_name
        if "-" in sequence_code:
            sequence_code, neg_offset = sequence_code.split("-")
            atom_name = f"{atom_name}-{neg_offset}"

        key = chain_code, sequence_code
        shift_value = shift.value

        residue_shifts.setdefault(key, {})[atom_name] = shift_value

    return residue_shifts


def _select_single_frame_or_exit(entry, name):

    frames = select_frames_by_name(entry.frame_list, name)

    if len(frames) == 0:
        exit_error(
            f"no frames selected from entry: {entry.entry_id} by selector {name}"
        )

    if len(frames) > 1:

        msg = f"""\
            multiple frames selected from entry: {entry.entry_id} by selector {name}
            frames selected were {' '.join([frame.name for frame in frames])}
        """
        exit_error(msg)

        exit_error(msg)

    return frames[0]


# def _find_best_match(shifts_1, residue_atom_shifts_2, ignored_atoms, amide_weight, amide_search_region):
#     matches = {}
#     for key, shifts_2 in residue_atom_shifts_2.items():
#         atom_names_1 = set(shifts_1.keys())
#         atom_names_2 = set(shifts_2.keys())
#
#         common_atom_names = atom_names_1.intersection(atom_names_2)
#
#
#
#         if len(common_atom_names) != 0:
#
#             differences = {}
#             for atom_name in common_atom_names:
#                 diff = abs(shifts_1[atom_name] - shifts_2[atom_name])
#                 if atom_name in 'H N'.split():
#                     diff = diff * amide_weight
#                 differences[atom_name] = diff
#
#             ok = True
#
#             for amide_name in amide_search_region:
#                 if amide_name in differences:
#                     if differences[amide_name] > amide_search_region[amide_name]:
#                         ok = False
#
#             if ok:
#                 for name in ignored_atoms:
#                     if name in matches:
#                         del matches[name]
#
#                 matches[sum(differences.values())] = key
#
#     return matches
#
#
# def _squeeze_spaces_left(string):
#     pre_len = len(string)
#     string = string.replace(' ', '')
#     post_len = len(string)
#     pad = " " * (pre_len - post_len)
#     string = f'{pad}{string}'
#
#     return string
#
#
#



import sys
from dataclasses import dataclass, replace
from pathlib import Path
from typing import List, Union

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    loop_row_namespace_iter,
    read_entry_from_file_or_exit_error,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (  # sequence_from_entry_or_exit,
    get_chain_code_iter,
)
from nef_pipelines.lib.structures import AtomLabel, RdcRestraint, Residue
from nef_pipelines.lib.util import STDIN, exit_error, parse_comma_separated_options
from nef_pipelines.transcoders.nmrstar import import_app
from nef_pipelines.transcoders.pales.importers.rdcs import _RDCS_to_nef_saveframe


@import_app.command()
def rdcs(
    chain_codes: List[str] = typer.Option(
        None,
        "--chains",
        help="chain codes as a list of names separated by commas, repeated calls will add further chains [default A]",
        metavar="<CHAIN-CODES>",
    ),
    frame_name: str = typer.Option(
        None,
        "-f",
        "--frame-name",
        help="a name for the frame, default is the name nmrstar entry",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="file to read NEF data from [- is stdin; defaults is stdin]]",
    ),
    use_author: bool = typer.Option(
        False, help="use author fields for chain_code, sequence_code and residue_name"
    ),
    file_path: Path = typer.Argument(
        ..., help="input files of type shifts.txt", metavar="<NMR-STAR-shifts>.str"
    ),
):
    """- convert nmrstar shift file <nmr-star>.str to NEF [alpha]"""

    nmrstar_entry = read_entry_from_file_or_exit_error(file_path)

    chain_codes = parse_comma_separated_options(chain_codes)

    nef_entry = read_entry_from_file_or_stdin_or_exit_error(input)

    nef_entry = pipe(
        nef_entry,
        chain_codes,
        frame_name,
        nmrstar_entry,
        file_path,
        use_author,
    )

    print(nef_entry)


def pipe(
    nef_entry: Entry,
    chain_codes: List[str],
    frame_name: str,
    nmrstar_entry: Entry,
    file_name: Path,
    use_author: bool,
):

    # sequence_residues = sequence_from_entry_or_exit(nef_entry)

    # sequence_residues_by_residue = {
    #     Residue.from_sequence_residue(sequence_residue): sequence_residue
    #     for sequence_residue in sequence_residues
    # }

    # residues = set(sequence_residues_by_residue.keys())

    rdc_lists, name_info = _rdcs_from_star_frame(nmrstar_entry, use_author, file_name)

    entity_id_to_chain_code = _build_entity_ids_to_chain_codes(chain_codes, rdc_lists)

    rdc_lists = _replace_shift_entity_ids_with_chain_codes(
        rdc_lists, entity_id_to_chain_code
    )

    for rdc_list, name_info in zip(rdc_lists, name_info):
        frame_name = name_info.entry_id + "_" + str(name_info.identifier).lstrip("_")
        rdcs_frame = _RDCS_to_nef_saveframe(rdc_list, frame_name)
        nef_entry.add_saveframe(rdcs_frame)

    return nef_entry


@dataclass
class ShiftListNameInfo:
    entry_id: str
    identifier: Union[int, str]


ELEMENT_TO_ISOTOPE = {"C": 13, "N": 15, "P": 31, "H": 1, "F": 19, "D": 2}


def _element_to_isotope(element):
    return ELEMENT_TO_ISOTOPE[element] if element in ELEMENT_TO_ISOTOPE else None


def _rdcs_from_star_frame(nmrstar_entry, use_author, file_name):

    raw_rdc_lists = nmrstar_entry.get_saveframes_by_category("RDC_constraints")

    if len(raw_rdc_lists) == 0:
        msg = f"""\
        there are no saveframes of assigned chemical shifts in the file {file_name}
        """

        exit_error(msg)

    # for i, raw_rdc_list in enumerate(raw_rdc_lists):
    #     category = raw_rdc_lists[0].get_tag("Sf_category")[0]
    #     entry_id = raw_rdc_lists[0].get_tag("Entry_ID")[0]
    #     frame_code = raw_rdc_lists[0].get_tag("Sf_framecode")[0]
    #     list_identifier = frame_code[len(category) + 1:].lstrip("_")
    #     name_info = ShiftListNameInfo(entry_id, list_identifier)
    #     print(category, entry_id, frame_code, list_identifier, name_info)

    name_info_list = []
    rdc_lists = []
    for i, raw_rdc_list in enumerate(raw_rdc_lists):
        category = raw_rdc_list.get_tag("Sf_category")[0]
        entry_id = raw_rdc_list.get_tag("Entry_ID")[0]
        frame_code = raw_rdc_list.get_tag("Sf_framecode")[0]
        list_identifier = frame_code[len(category) + 1 :].lstrip("_")
        name_info = ShiftListNameInfo(entry_id, list_identifier)
        name_info_list.append(name_info)

        try:
            rdc_loop = raw_rdc_list.get_loop("_RDC_constraint")

        except KeyError:
            msg = f"""\
                warning the saveframe {raw_rdc_list.get_tag('Sf_framecode')[0]}
                does not contain a loop with the tag _Atom_chem_shift
                """
            print(msg, file=sys.stderr)
            continue

        # TODO add warnings and check all required fields are present
        rdc_list = []
        rdc_lists.append(rdc_list)
        for row in loop_row_namespace_iter(rdc_loop):

            chain_code_1 = row.Entity_assembly_ID_1
            chain_code_2 = row.Entity_assembly_ID_2

            if use_author:
                atom_id_1 = row.Auth_atom_ID_1
                seq_id_1 = row.Auth_seq_ID_1
                comp_id_1 = row.Auth_comp_ID_1

                atom_id_2 = row.Auth_atom_ID_2
                seq_id_2 = row.Auth_seq_ID_2
                comp_id_2 = row.Auth_comp_ID_2
            else:
                atom_id_1 = row.Atom_ID_1
                seq_id_1 = row.Seq_ID_1
                comp_id_1 = row.Comp_ID_1

                atom_id_2 = row.Atom_ID_2
                seq_id_2 = row.Seq_ID_2
                comp_id_2 = row.Comp_ID_2

            if atom_id_1 == UNUSED and use_author:
                atom_id_1 = row.Atom_ID_1
            if atom_id_2 == UNUSED and use_author:
                atom_id_2 = row.Atom_ID_2

            if seq_id_1 == UNUSED and use_author:
                seq_id_1 = row.Seq_ID_1
            if seq_id_2 == UNUSED and use_author:
                seq_id_2 = row.Seq_ID_2

            if comp_id_1 == UNUSED and use_author:
                comp_id_1 = row.Comp_ID_1
            if comp_id_2 == UNUSED and use_author:
                comp_id_2 = row.Comp_ID_2

            val = row.RDC_val
            sdev = row.RDC_val_err

            lower_bound = row.RDC_lower_bound
            upper_bound = row.RDC_upper_bound
            if val == UNUSED:
                val = (lower_bound + upper_bound) / 2
                val_range_d2 = (upper_bound - lower_bound) / 2
                if sdev < val_range_d2 or sdev == UNUSED:
                    sdev = val_range_d2

            weight = (
                row.RDC_val_scale_factor if row.RDC_val_scale_factor != UNUSED else None
            )

            element_1 = row.Atom_type_1
            element_2 = row.Atom_type_2
            isotope_1 = (
                row.Atom_isotope_number_1
                if row.Atom_isotope_number_1 != UNUSED
                else _element_to_isotope(element_1)
            )
            isotope_2 = (
                row.Atom_isotope_number_2
                if row.Atom_isotope_number_2 != UNUSED
                else _element_to_isotope(element_2)
            )

            residue_1 = Residue(chain_code_1, seq_id_1, comp_id_1)
            atom_1 = AtomLabel(residue_1, atom_id_1, element_1, isotope_1)

            residue_2 = Residue(chain_code_2, seq_id_2, comp_id_2)
            atom_2 = AtomLabel(residue_2, atom_id_2, element_2, isotope_2)

            if upper_bound == UNUSED and lower_bound == UNUSED:
                rdc = RdcRestraint(atom_1, atom_2, val, sdev)
            else:
                rdc = RdcRestraint(
                    atom_1,
                    atom_2,
                    val,
                    sdev,
                    lower_bound=lower_bound,
                    upper_bound=upper_bound,
                    weight=weight,
                )
            rdc_list.append(rdc)

    return rdc_lists, name_info_list


#
def _replace_shift_entity_ids_with_chain_codes(rdc_lists, entity_id_to_chain_code):

    for list_index, rdc_list in enumerate(rdc_lists):
        for rdc_index, rdc_restraint in enumerate(rdc_list):
            atom_1 = rdc_restraint.atom_1
            residue_1 = atom_1.residue
            residue_1 = replace(
                residue_1, chain_code=entity_id_to_chain_code[residue_1.chain_code]
            )
            atom_1 = replace(atom_1, residue=residue_1)

            atom_2 = rdc_restraint.atom_2
            residue_2 = atom_2.residue
            residue_2 = replace(
                residue_2, chain_code=entity_id_to_chain_code[residue_2.chain_code]
            )
            atom_2 = replace(atom_2, residue=residue_2)

            rdc_restraint = replace(rdc_restraint, atom_1=atom_1)
            rdc_restraint = replace(rdc_restraint, atom_2=atom_2)
            rdc_list[rdc_index] = rdc_restraint

    return rdc_lists


def _build_entity_ids_to_chain_codes(chain_codes, rdc_lists: List[RdcRestraint]):
    all_entity_ids_to_chain_codes = {}
    for rdc_list in rdc_lists:

        entity_ids_1 = OrderedSet(
            sorted([rdc.atom_1.residue.chain_code for rdc in rdc_list])
        )
        entity_ids_2 = OrderedSet(
            sorted([rdc.atom_2.residue.chain_code for rdc in rdc_list])
        )
        entity_ids = entity_ids_1.union(entity_ids_2)

        chain_code_iter = get_chain_code_iter(chain_codes)
        entity_ids_to_chain_codes = {
            entity_id: chain_code
            for entity_id, chain_code in zip(entity_ids, chain_code_iter)
        }
        all_entity_ids_to_chain_codes.update(entity_ids_to_chain_codes)
    return entity_ids_to_chain_codes



import typer

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    isolate_loop,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.sparky.importers.peaks import peaks

app = typer.Typer()
app.command()(peaks)


EXPECTED = """
loop_
  _nef_peak.index
  _nef_peak.peak_id
  _nef_peak.chain_code_1
  _nef_peak.sequence_code_1
  _nef_peak.residue_name_1
  _nef_peak.atom_name_1
  _nef_peak.chain_code_2
  _nef_peak.sequence_code_2
  _nef_peak.residue_name_2
  _nef_peak.atom_name_2
  _nef_peak.position_1
  _nef_peak.position_uncertainty_1
  _nef_peak.position_2
  _nef_peak.position_uncertainty_2
  _nef_peak.height
  _nef_peak.height_uncertainty
  _nef_peak.volume
  _nef_peak.volume_uncertainty


     1   1   A   16   DG   H3'   A   16   DG   H8    4.905   .   8.01    .   .   .   7150000.0    .
     2   2   A   16   DG   H4'   A   16   DG   H8    4.439   .   8.013   .   .   .   5420000.0    .
     3   3   A   17   DT   H6    A   16   DG   H8    7.205   .   8.004   .   .   .   1680000.0    .
     4   4   A   17   DT   H7    A   16   DG   H8    1.459   .   8.008   .   .   .   20900000.0   .
     5   5   A   17   DT   H2"   A   17   DT   H1'   2.509   .   5.84    .   .   .   46800000.0   .

stop_
"""

DATA_SEQUENCE = read_test_data("sparky_manual_basic_sequence.nef", __file__)


def test_basic():

    path = path_in_test_data(__file__, "sparky_manual_basic.peaks")

    result = run_and_report(app, ["--molecule-type", "dna", path], input=DATA_SEQUENCE)

    loop = isolate_loop(
        result.stdout, "nef_nmr_spectrum_sparky_sparky_manual_basic", "nef_peak"
    )

    assert_lines_match(EXPECTED, str(loop))


def test_basic_no_sequence():

    path = path_in_test_data(__file__, "sparky_manual_basic.peaks")

    result = run_and_report(app, ["--molecule-type", "dna", path])

    loop = isolate_loop(
        result.stdout, "nef_nmr_spectrum_sparky_sparky_manual_basic", "nef_peak"
    )

    assert_lines_match(EXPECTED, str(loop))


def test_basic_no_sequence_requires_sequence():

    path = path_in_test_data(__file__, "sparky_manual_full_no_sequence.peaks")

    result = run_and_report(app, [path], input=DATA_SEQUENCE)

    loop = isolate_loop(
        result.stdout,
        "nef_nmr_spectrum_sparky_sparky_manual_full_no_sequence",
        "nef_peak",
    )

    assert_lines_match(EXPECTED, str(loop))


def test_full():

    path = path_in_test_data(__file__, "sparky_manual_full.peaks")

    result = run_and_report(app, ["--molecule-type", "dna", path])

    loop = isolate_loop(
        result.stdout, "nef_nmr_spectrum_sparky_sparky_manual_full", "nef_peak"
    )

    assert_lines_match(EXPECTED, str(loop))


EXPECTED_FULL_COMMENT = """ # noqa: E501
loop_
  _nef_peak.index
  _nef_peak.peak_id
  _nef_peak.chain_code_1
  _nef_peak.sequence_code_1
  _nef_peak.residue_name_1
  _nef_peak.atom_name_1
  _nef_peak.chain_code_2
  _nef_peak.sequence_code_2
  _nef_peak.residue_name_2
  _nef_peak.atom_name_2
  _nef_peak.position_1
  _nef_peak.position_uncertainty_1
  _nef_peak.position_2
  _nef_peak.position_uncertainty_2
  _nef_peak.height
  _nef_peak.height_uncertainty
  _nef_peak.volume
  _nef_peak.volume_uncertainty
  _nef_peak.ccpn_comment

     1   1   A   16   DG   H3'   A   16   DG   H8    4.905   .   8.01    .   .   .   7150000.0    .   'this is a comment'
     2   2   A   16   DG   H4'   A   16   DG   H8    4.439   .   8.013   .   .   .   5420000.0    .   .
     3   3   A   17   DT   H6    A   16   DG   H8    7.205   .   8.004   .   .   .   1680000.0    .   'so is this as well'
     4   4   A   17   DT   H7    A   16   DG   H8    1.459   .   8.008   .   .   .   20900000.0   .   .
     5   5   A   17   DT   H2"   A   17   DT   H1'   2.509   .   5.84    .   .   .   46800000.0   .   .
stop_
""".replace(
    NOQA_E501, ""
)


def test_full_comment():

    path = path_in_test_data(__file__, "sparky_manual_full_comment.peaks")

    result = run_and_report(app, ["--molecule-type", "dna", path])

    loop = isolate_loop(
        result.stdout, "nef_nmr_spectrum_sparky_sparky_manual_full_comment", "nef_peak"
    )

    assert_lines_match(EXPECTED_FULL_COMMENT, str(loop))


def test_missing_isotopes():
    path = path_in_test_data(__file__, "CONCACX.txt")

    result = run_and_report(
        app,
        [
            path,
        ],
        expected_exit_code=1,
    )

    assert "the isotopes are not defined" in result.stdout
    assert "[1, 2, 3, 4]" in result.stdout


EXPECTED_SPECTRUM_CONCACX = open(
    path_in_test_data(__file__, "CONCACX_expected_spectrum.txt")
).read()


def test_give_isotopes_when_bad():
    path = path_in_test_data(__file__, "CONCACX.txt")

    result = run_and_report(
        app,
        [
            "--nuclei",
            "13C,15N,13C,13C",
            path,
        ],
    )

    spectrum_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_sparky_CONCACX")

    assert_lines_match(EXPECTED_SPECTRUM_CONCACX, spectrum_frame)



# original implementation by esther
import sys
from pathlib import Path

import typer
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import read_entry_from_stdin_or_exit
from nef_pipelines.lib.util import STDOUT, exit_if_file_has_bytes_and_no_force
from nef_pipelines.transcoders.mars import export_app

MARS_DEFAULT_FILE = "mars.inp"

INPUT_TEMPLATE = """\
     fragSize:     5                      # Maximum length of pseudoresidue fragments

     cutoffCO:     0.25                   # Connectivity cutoff (ppm) of CO [0.25]
     cutoffCA:     0.2                    # Connectivity cutoff (ppm) of CA [0.5]
     cutoffCB:     0.5                    # Connectivity cutoff (ppm) of CB [0.5]
     cutoffHA:     0.25                   # Connectivity cutoff (ppm) of HA [0.25]
     cutoffHN:     0.02                   # Connectivity cutoff (ppm) of HN [0.15]
     cutoffN:      0.05                   # Connectivity cutoff (ppm) of N [0.10]
     cutoffnHN:    0.02                   # Connectivity cutoff (ppm) of HN+1 [0.15]
     cutoffnN:     0.05                   # Connectivity cutoff (ppm) of N+1 [0.10]

     fixConn:      {root}_fix_con.tab     # Table for fixing sequential connectivity
     fixAss:       {root}_fix_ass.tab     # Table for fixing residue type and(or) assignment

     pdb:          0                      # 3D structure available [0/1]
     resolution:   NO                     # Resolution of 3D structure [Angstrom]
     pdbName:      NO                     # Name of PDB file (protons required!)
     tensor:       NO                     # Method for obtaining alignment tensor [0/1/2/3/4]

     nIter:        NO                     # Number of iterations [2/3/4]

     dObsExh:      NO                     # Name of RDC table for exhaustive SVD (PALES format)
     dcTab:        NO                     # Name of RDC table (PALES format)

     deuterated:   {deuterated}           # Protonated proteins [0]; perdeuterated proteins [1]

     rand_coil:    {random_coil}          # Folded proteins [0]; disordered proteins [1]

     sequence:     {root}.fasta           # Primary sequence (FASTA format)
     secondary:    {root}_psipred.tab     # Secondary structure (PSIPRED format)
     csTab:        {root}_shifts.tab      # Chemical shift table
 """


# noinspection PyUnusedLocal
@export_app.command("input")
def input_(
    deuterated: bool = typer.Option(
        False,
        help="is the protein deuterated [default: False]",
        metavar="<DEUTERATED-PROTEIN>",
    ),
    random_coil: bool = typer.Option(
        False, help="is the protein folded [default: False]", metavar="<FOLDED-PROTEIN>"
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
    output_path: Path = typer.Argument(
        None,
        help="directory file name to output to [default <entry_id>.inp] for stdout use -",
        metavar="<MARS-INPUT-FILENAME>",
    ),
):
    """- write mars input fixed assignment and connectivity files"""

    entry = read_entry_from_stdin_or_exit()

    output_file = Path(output_path) if output_path else Path(f"{entry.entry_id}.inp")

    entry = pipe(entry, deuterated, random_coil, output_file, force)

    if entry:
        print(entry)


def pipe(
    entry: Entry, deuterated: bool, random_coil: bool, output_path: Path, force: bool
) -> Entry:

    deuterated = int(deuterated)
    random_coil = int(random_coil)

    output_path = (
        output_path / f"{entry.entry_id}.inp" if output_path.is_dir() else output_path
    )

    root = (
        output_path.parent / output_path.stem
        if output_path != STDOUT
        else f"{entry.entry_id}"
    )

    templated_text = INPUT_TEMPLATE.format(
        root=root, deuterated=deuterated, random_coil=random_coil
    )

    exit_if_file_has_bytes_and_no_force(output_path, force)

    file_h = sys.stdout if output_path == STDOUT else open(output_path, "w")

    print(templated_text, file=file_h)

    if output_path != STDOUT:
        file_h.close()

    if output_path != STDOUT:
        Path(f"{root}_fix_con.tab").touch()
        Path(f"{root}_fix_ass.tab").touch()

    result = entry if output_path != STDOUT else None

    return result



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()

if nef_app.app:
    nef_app.app.add_typer(app, name="deep", help="- read deep [peaks]")

    app.add_typer(import_app, name="import", help="- import deep [peaks]")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.deep.importers.peaks  # noqa: F401



# TODO: xplor names -> iupac and deal with %% and ## properly
# TODO: add common experiment types
# TODO: add a chemical shift list reference
# TODO: _nef_nmr_spectrum: value_first_point, folding, absolute_peak_positions, is_acquisition
# TODO: cleanup
# TODO: multiple assignments per peak... howto in nef
# TODO: add libs pipeline

import itertools
import string
import sys
import traceback
from collections import Counter, OrderedDict
from dataclasses import dataclass, replace
from enum import auto
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Tuple

import typer
from fyeah import f
from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib import constants
from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.isotope_lib import GAMMA_RATIOS
from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    MoleculeType,
    get_chain_starts_and_ends,
    get_residue_name_from_lookup,
    sequence_from_entry,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import (
    CCPN_UNSASSIGNED_CHAIN,
    PSEUDO_PREFIX,
    UNUSED,
    AtomLabel,
    LineInfo,
    PeakAxis,
    PeakList,
    PeakListData,
    PeakValues,
    SequenceResidue,
)
from nef_pipelines.lib.util import (
    NEWLINE,
    STDIN,
    exit_error,
    is_float,
    is_int,
    parse_comma_separated_options,
    strip_characters_left,
    warn,
)
from nef_pipelines.transcoders.nmrview import import_app

from ..nmrview_lib import parse_float_list, parse_tcl

app = typer.Typer()


class ResidueNameHandlingOption(LowercaseStrEnum):

    STOP = auto()  # stop processing
    PSEUDO = (
        auto()
    )  # continue processing but the residue becomes a pseudo residue with sequence code @xxx
    SEQUENCE = auto()  # use the residue type in the read sequence instead
    UNASSIGN = (
        auto()
    )  # set the chain_code, sequence code and  residue type to . and continue
    COMMENT = (
        auto()
    )  # save the assignment information for the peak as a comment in the NEF file
    CONTINUE = auto()  # don't change anything take whats in the file

    CASE = (
        auto()
    )  # if the residue name matches the sequence but the case is different correct it
    WARN = auto()  # report the peak with missing residue information


class ResidueNumberHandlingOption(LowercaseStrEnum):

    STOP = auto()  # stop processing
    PSEUDO = (
        auto()
    )  # continue processing but the residue becomes a pseudo residue with sequence code @xxx
    UNASSIGN = (
        auto()
    )  # set the chain_code, sequence code and  residue type to . and continue
    COMMENT = (
        auto()
    )  # save the assignment information for the peak in a comment field in NEF
    CONTINUE = auto()  # don't change anything, take whats in the file

    WARN = auto()  # report each peak with missing residue information
    GLOBAL_WARN = (
        auto()
    )  # issue an overall warning that there are residues outside the sequence


PSEUDO_OPTION_INFO = """\
   continue processing but the residue becomes a pseudo residue with its sequence code replaced by a pseudo residue
   (xxx -> @xxx)
"""
UNASSIGN_OPTION_INFO = "remove the assignment: set the chain_code, sequence code and  residue type to . and continue"

RESIDUE_NUMBER_HANDLING_OPTION_EXPLANATIONS = {
    ResidueNumberHandlingOption.STOP: "stop processing",
    ResidueNumberHandlingOption.PSEUDO: PSEUDO_OPTION_INFO,
    ResidueNumberHandlingOption.UNASSIGN: UNASSIGN_OPTION_INFO,
    ResidueNumberHandlingOption.COMMENT: "place the information about the assignment in a comment and unassign it",
    ResidueNumberHandlingOption.CONTINUE: "don't change anything, take whats in the file",
    ResidueNumberHandlingOption.WARN: "report the peak with missing residue information",
}
RESIDUE_NAME_HANDLING_HELP = """
what to do when the residue name for an assignment isn't defined in the sequence (if read) but it has a chain and
residue number and the residue name doesn't match the one in the sequence. Choices are
stop: stop processing if the residue name isn't defined or is mismatched
pseudo: continue processing but the residue becomes a pseudo residue with sequence_code @xxx and the residue name read
sequence: continue processing but the residue becomes the residue name read from the sequence
unassign: set the chain_code, sequence code and residue type to . and continue without warning
comment: put all the assignment information in the peak comment and continue
continue: leave everything it as it is

two modifiers can be added to the choice by adding them as extra options

warn: report the peak with mismatched residue information on STDERR, use the read values and continue processing
case: if the residue name matches the sequence but the case is different correct it and continue
"""

RESIDUE_NUMBER_HANDLING_HELP = """
what to do when the residue number for an assignment isn't defined in the sequence. Choices are:
stop: stop processing if the residue name isn't defined or is mismatched
pseudo: continue processing but the residue becomes a pseudo residue with sequence_code @xxx
unassign: set the chain_code, sequence code and residue type to . and continue without warning
comment: put all the assignment information in the peak comment and continue
continue: leave everything it as it is

one modifiers can be added to the choice by adding it as an extra option

warn: report the peak with mismatched residue information on STDERR, use the read values and continue processing
"""


class ResidueNameTypeOption(LowercaseStrEnum):
    AUTO = auto()
    SINGLE = auto()
    THREE = auto()


RESIDUE_NAME_TYPE_HELP = """
auto: translate name based on its length otherwise, single: single letter amino acid code
three: three letter or longer amino acid code
"""


class FrameNameOption(LowercaseStrEnum):
    SPECTRUM = auto()
    FILE = auto()


FRAME_NAME_SOURCE_HELP = """ where to get the frame name from"""


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    in_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    chain_codes: List[str] = typer.Option(
        None, "--chain", help="chain code", metavar="<chain-code>"
    ),
    residue_name_handling: List[str] = typer.Option(
        None, "--residue-name-handling", help=RESIDUE_NAME_HANDLING_HELP
    ),
    residue_number_handling: List[str] = typer.Option(
        None, "--residue-number-handling", help=RESIDUE_NUMBER_HANDLING_HELP
    ),
    residue_name_type: ResidueNameTypeOption = typer.Option(
        ResidueNameTypeOption.AUTO, "--residue-name-type", help=RESIDUE_NAME_TYPE_HELP
    ),
    molecule_type: MoleculeType = typer.Option(
        MoleculeType.PROTEIN,
        "--molecule-type",
        help="molecule type one of protein, dna, rna or  carbohydrate",
    ),
    entry_name: str = typer.Option(
        "nmrview", "-n", "--name", help="entry name", metavar="<entry-name>"
    ),
    frame_name_source: FrameNameOption = typer.Option(
        FrameNameOption.SPECTRUM, help=FRAME_NAME_SOURCE_HELP
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert NMRView peak file <nmrview>.xpk files to NEF [alpha for new residue name and sequence code handling]"""

    if not chain_codes:
        chain_codes = ["A"] * len(file_names)

    if len(chain_codes) == 1:
        chain_codes = chain_codes * len(file_names)

    chain_codes = parse_comma_separated_options(chain_codes)

    entry = read_or_create_entry_exit_error_on_bad_file(in_file, entry_name)

    sequence = sequence_from_entry(entry)

    _exit_if_num_chains_and_files_dont_match(chain_codes, file_names)

    residue_name_handling = parse_comma_separated_options(residue_name_handling)

    residue_name_handling = _fixup_residue_handling_or_exit_bad(residue_name_handling)

    residue_number_handling = parse_comma_separated_options(residue_number_handling)

    residue_number_handling = _fixup_residue_number_handling_or_exit_bad(
        residue_number_handling
    )

    entry = pipe(
        entry,
        file_names,
        chain_codes,
        sequence,
        molecule_type,
        residue_name_type,
        residue_name_handling,
        residue_number_handling,
        frame_name_source,
    )

    print(entry)


def pipe(
    entry: Entry,
    file_names: List[Path],
    chain_codes: List[str],
    sequence: List[SequenceResidue],
    molecule_type: MoleculeType,
    residue_name_type: ResidueNameTypeOption,
    residue_name_handling: ResidueNameHandlingOption,
    residue_number_handling: ResidueNumberHandlingOption,
    frame_name_source: FrameNameOption,
) -> Entry:

    frames = []

    frame_names_and_peak_lists = []
    for file_name, chain_code in zip(file_names, chain_codes):
        with open(file_name, "r") as lines:
            peaks_list = _read_raw_peaks(
                lines,
                chain_code,
                sequence,
                file_name,
                molecule_type,
                residue_name_type,
                residue_name_handling,
                residue_number_handling,
            )

        frame_name = _make_peak_list_frame_name(
            peaks_list, file_name, frame_name_source
        )

        frame_names_and_peak_lists.append((frame_name, peaks_list))

    frame_names_and_peak_lists = _disambiguate_frame_names(frame_names_and_peak_lists)

    for (frame_name, peaks_list), chain_code in zip(
        frame_names_and_peak_lists, chain_codes
    ):
        frames.append(_create_spectrum_frame(frame_name, peaks_list, chain_code))

    return add_frames_to_entry(entry, frames)


def _sequence_code_out_of_range(peak, axis, chain_starts_and_ends):
    atom_labels = peak[axis].atom_labels
    sequence_code = atom_labels.residue.sequence_code
    chain_code = atom_labels.residue.chain_code
    chain_start_and_end = chain_starts_and_ends[chain_code]

    if is_int(sequence_code):
        out_of_range = (
            sequence_code < chain_start_and_end[0]
            or sequence_code > chain_start_and_end[-1]
        )
    else:
        out_of_range = False

    return out_of_range


def _fix_up_peak_labels(peak, axis, residue_number_handling):

    atom_labels = peak[axis].atom_labels
    residue = atom_labels.residue

    if ResidueNumberHandlingOption.PSEUDO in residue_number_handling:
        residue = replace(
            residue,
            sequence_code_prefix=PSEUDO_PREFIX,
            chain_code_prefix=PSEUDO_PREFIX,
            chain_code=CCPN_UNSASSIGNED_CHAIN,
        )
        atom_labels = replace(atom_labels, residue=residue)
    elif ResidueNumberHandlingOption.UNASSIGN in residue_number_handling:
        residue = replace(
            residue,
            chain_code=UNUSED,
            sequence_code=UNUSED,
            residue_name=UNUSED,
            sequence_code_prefix="",
            offset=0,
        )
        atom_labels = replace(atom_labels, residue=residue)
        atom_labels = replace(atom_labels, atom_name=UNUSED)

    peak[axis] = replace(peak[axis], atom_labels=atom_labels)
    return peak


def _copy_assignment_to_comments(peak, axis, peak_number_handling):

    atom_labels = peak[axis].atom_labels
    peak_comment = peak["values"].comment

    WARNING = "WARNING: some residue sequence codes are out of range:"
    if ResidueNumberHandlingOption.COMMENT in peak_number_handling:
        if WARNING not in peak_comment:
            peak_comment = f"{WARNING}\n{peak_comment}"
        residue = atom_labels.residue
        atom_name = atom_labels.atom_name if atom_labels.atom_name != UNUSED else ""
        sequence_code = residue.sequence_code if residue.sequence_code != UNUSED else ""
        chain_code = residue.chain_code if residue.chain_code != UNUSED else ""
        residue_name = residue.residue_name if residue.residue_name != UNUSED else ""

        assignment_items = [chain_code, sequence_code, residue_name, atom_name]
        assignment_items = [str(item) for item in assignment_items]
        assignment = ".".join(assignment_items)
        peak_comment = f"{peak_comment}\nassignment with sequence_code out of range [axis {axis+1}]: {assignment}"

        peak["values"] = replace(peak["values"], comment=peak_comment)

    return peak


def _handle_bad_sequence_codes(
    peak, chain_starts_and_ends, peak_number_handling, line_info
):

    for item_id in peak:
        # this is because we are using dicts use a dataclass!
        if not isinstance(item_id, int):
            continue
        axis = item_id

        # TODO we should be iterating over peak labels
        chain_code = peak[axis].atom_labels.residue.chain_code

        sequence_out_of_range = False
        if chain_code in chain_starts_and_ends:
            chain_start_and_end = chain_starts_and_ends[chain_code]

            if sequence_out_of_range := _sequence_code_out_of_range(
                peak, axis, chain_starts_and_ends
            ):

                _warn_or_exit_for_sequence_code_out_of_range(
                    peak, axis, chain_start_and_end, line_info, peak_number_handling
                )

                if ResidueNumberHandlingOption.CONTINUE not in peak_number_handling:
                    peak = _copy_assignment_to_comments(
                        peak, axis, peak_number_handling
                    )
                    peak = _fix_up_peak_labels(peak, axis, peak_number_handling)

            # else:
            # its just a string leave it as it is...

    return sequence_out_of_range


def _warn_or_exit_for_sequence_code_out_of_range(
    peak, axis, chain_start_and_end, line_info, peak_number_handling
):
    atom_labels = peak[axis].atom_labels
    chain_code = atom_labels.residue.chain_code
    sequence_code = atom_labels.residue.sequence_code

    if (do_warn := ResidueNumberHandlingOption.WARN in peak_number_handling) or (
        do_stop := ResidueNumberHandlingOption.STOP in peak_number_handling
    ):

        if do_warn:
            handling_options = set(peak_number_handling)
            if ResidueNumberHandlingOption.WARN in handling_options:
                handling_options.remove(ResidueNumberHandlingOption.WARN)
            action_taken = list(handling_options)[0]
            action = f"action taken: {RESIDUE_NUMBER_HANDLING_OPTION_EXPLANATIONS[action_taken]}"
        elif do_stop:
            action = ""

        msg = f"""\
                at line number {line_info.line_no} in file {line_info.file_name} with the value:

                '{line_info.line.strip()}'

                on axis {axis + 1} the residue number {sequence_code} is outside the bounds of it's
                chain {chain_code} [{chain_start_and_end[0]} - {chain_start_and_end[1]}]
                {action}
              """

        if do_warn:
            warn(msg)
        elif do_stop:
            exit_error(msg)


def _disambiguate_frame_names(peak_lists_and_entry_names):
    seen_frames = Counter()
    for i, (frame_name, peaks_list) in enumerate(peak_lists_and_entry_names):
        if frame_name in seen_frames:
            seen_frames[frame_name] += 1
            frame_name = f"{frame_name}`{seen_frames[frame_name]}`"
            peak_lists_and_entry_names[i] = (frame_name, peaks_list)
        else:
            seen_frames[frame_name] = 1
    return peak_lists_and_entry_names


def _create_entry_names_from_template_if_required(
    entry_name_template, entry_names, file_names
):
    new_entry_names = []
    for file_name, entry_name in itertools.zip_longest(file_names, entry_names):
        if entry_name:
            new_entry_names.append(entry_name)
        else:
            file_name = str(Path(file_name).stem)
            numbers_and_letters = string.ascii_lowercase + string.digits
            file_name = "".join(
                [
                    letter if letter in numbers_and_letters else "_"
                    for letter in file_name
                ]
            )
            file_name = f"_{file_name}"
            new_entry_names.append(f(entry_name_template))
    return new_entry_names


def _read_raw_peaks(
    lines,
    chain_code,
    sequence,
    file_name,
    molecule_type,
    residue_name_type,
    residue_name_handling,
    residue_number_handling,
):

    header = get_header_or_exit(lines)

    header_data = read_header_data(lines, header)

    column_indices = read_peak_columns(lines, header_data)

    raw_peaks = _read_peak_data(
        lines,
        header_data,
        column_indices,
        chain_code,
        sequence,
        file_name,
        molecule_type,
        residue_name_type,
        residue_name_handling,
        residue_number_handling,
    )

    return PeakList(header_data, raw_peaks)


def _dedent_all(text):
    lines = []
    for line in text.split("\n"):
        lines.append(line.lstrip())

    return "\n".join(lines)


def _read_peak_data(
    lines,
    header_data,
    column_indices,
    chain_code,
    sequence,
    file_name,
    molecule_type,
    residue_name_type,
    residue_name_handling,
    residue_number_handling,
):
    sequence_lookup = _sequence_to_residue_type_lookup(sequence)
    chain_starts_and_ends = get_chain_starts_and_ends(sequence)

    raw_peaks = []
    field = None

    bad_peak_serials = []

    for line_no, raw_line in enumerate(lines):
        line_info = LineInfo(file_name, line_no, raw_line)
        if not len(raw_line.strip()):
            continue
        try:
            peak = {}
            tcl_list = parse_tcl(raw_line)
            # TODO validate and report errors
            peak_serial = int(tcl_list[0])

            for axis_index, axis in enumerate(header_data.axis_labels):
                axis_values = _read_axis_for_peak(
                    tcl_list,
                    axis,
                    column_indices,
                    chain_code,
                    sequence_lookup,
                    molecule_type,
                    residue_name_type,
                    residue_name_handling,
                    line_info,
                )

                peak[axis_index] = PeakAxis(*axis_values)

            raw_values = _read_values_for_peak(tcl_list, column_indices)
            peak["values"] = PeakValues(peak_serial, **raw_values)

            sequence_code_out_of_range = _handle_bad_sequence_codes(
                peak, chain_starts_and_ends, residue_number_handling, line_info
            )

            if sequence_code_out_of_range:
                bad_peak_serials.append(peak_serial)

            raw_peaks.append(peak)

        except Exception as e:
            _report_parsing_exception_and_exit(e, axis_index, field, line_info)

    if ResidueNumberHandlingOption.GLOBAL_WARN in residue_number_handling:
        _globally_report_bad_sequence_code_if_present(bad_peak_serials, file_name)

    return raw_peaks


def _report_parsing_exception_and_exit(e, axis_index, field, line_info):
    field = str(field) if field else "unknown"
    traceback_text = traceback.format_exc()
    msg_text = f"""\
        There was an unexpected error parsing the NMRView Peak (xpk) file data
        at line {line_info.line_no} in the file {line_info.file_name}
        with input: '{line_info.line.strip()}'
        field: {field}  axis: {axis_index + 1}
        reason: {e}
        please contact the developers with your data file and this message
    """
    msg_text = dedent(msg_text).strip()
    msg = f"""{traceback_text}\n{msg_text}"""
    exit_error(msg)


def _globally_report_bad_sequence_code_if_present(bad_peak_serials, file_name):
    if bad_peak_serials:
        bad_peak_serials = [
            str(bad_peak_serial) for bad_peak_serial in bad_peak_serials
        ]
        bad_peak_serials_msg = ", ".join(bad_peak_serials)
        msg = f"""
                For the NMRView peak list file [xpk]: {file_name}
                some sequence codes didn't match the read molecular system, see comments on peaks in the NEF stream.
                The NMRView peak serials were

                {bad_peak_serials_msg}

                to hide this message and choose processing options use
                    nmrview import peaks --peak-number-handling

                note: typical values are stop, warn, comment, pseudo, and unassign which maybe be combined as a commma
                      separated list
            """
        msg = dedent(msg).strip()
        warn(msg)


def _exit_if_num_chains_and_files_dont_match(chain_codes, file_names):
    if len(chain_codes) != len(file_names):
        msg = f"""
            the number of chain codes {len(chain_codes)} does not match number of files {len(file_names)}
            the chain codes are {', '.join(chain_codes)}
            the files are
            {NEWLINE.join(file_names)}
        """
        exit_error(msg)


def read_peak_columns(lines, header_data):
    line = next(lines)
    raw_headings = line.split()
    heading_indices = OrderedDict({"index": 0})
    for axis_index, axis in enumerate(header_data.axis_labels):
        for axis_field in list("LPWBEJU"):
            header = f"{axis}.{axis_field}"
            if header in raw_headings:
                heading_indices[header] = raw_headings.index(header) + 1
    for peak_item in ["vol", "int", "stat", "comment", "flag0"]:
        if peak_item in raw_headings:
            heading_indices[peak_item] = raw_headings.index(peak_item) + 1
    return heading_indices


def read_header_data(lines, headers):
    data_set = None
    sweep_widths = []
    spectrometer_frequencies = []
    num_axis = None
    axis_labels = None
    for header_no, header_type in enumerate(headers):
        line = next(lines)
        if header_type == "label":
            axis_labels = line.strip().split()
            num_axis = len(axis_labels)
        elif header_type == "dataset":
            data_set = line.strip()
        elif header_type == "sw":
            line_no = header_no + 2
            sweep_widths = parse_float_list(line, line_no)
            check_num_fields(sweep_widths, num_axis, "sweep widths", line, line_no)
        elif header_type == "sf":
            line_no = header_no + 2
            spectrometer_frequencies = parse_float_list(line, line_no)
            check_num_fields(
                spectrometer_frequencies,
                num_axis,
                "spectrometer frequencies",
                line,
                line_no,
            )

    # sweep widths are in ppm for nef!
    sweep_widths = [float(sweep_width) for sweep_width in sweep_widths]
    spectrometer_frequencies = [
        float(spectrometer_frequency)
        for spectrometer_frequency in spectrometer_frequencies
    ]

    for i, (sweep_width, spectrometer_frequency) in enumerate(
        zip(sweep_widths, spectrometer_frequencies)
    ):
        sweep_widths[i] = f"{sweep_width / spectrometer_frequency:.4f}"

    # TODO: peaks shifts, spectrometer frequencies how many decimal points
    spectrometer_frequencies = [
        f"{spectrometer_frequency:4}"
        for spectrometer_frequency in spectrometer_frequencies
    ]

    peak_list_data = PeakListData(
        num_axis, axis_labels, data_set, sweep_widths, spectrometer_frequencies
    )
    return peak_list_data


def get_header_or_exit(lines):
    header_items = ["label", "dataset", "sw", "sf"]

    line = next(lines)

    headers = []
    if line:
        headers = line.strip().split()

    if len(headers) != 4:
        msg = f"""this doesn't look like an nmrview xpk file,
                  i expected a header containing 4 items on the first line: {','.join(header_items)}
                  i got {line} at line 1"""
        exit_error(msg)

    for name in header_items:
        if name not in headers:
            msg = f"""this doesn't look like an nmrview xpk file,
                       i expected a header containing the values: {', '.join(header_items)}
                       i got '{line}' at line 1"""
            exit_error(msg)

    return headers


def _exit_if_nmrview_residue_name_doesnt_match_sequence(
    nmrview_residue_name, chain_code, residue_number, residue_type, line_info
):

    if (
        residue_type != NEF_UNKNOWN
        and nmrview_residue_name
        and nmrview_residue_name != residue_type
    ):
        msg = f"""
                in file {line_info.file_name} at line {line_info.line_no}
                for residue number {residue_number} in chain {chain_code} the residue type in the NEF sequence
                [{residue_type}] does not match the one in the nmrview file [{nmrview_residue_name}]
                the line was
                 {line_info.line}

                to ignore this error run with --ignore-bad-residues which which will use residue names from the NEF
                sequence
            """
        exit_error(msg)


def _read_axis_for_peak(
    tcl_list,
    axis,
    heading_indices,
    chain_code,
    sequence_lookup,
    molecule_type,
    residue_name_type,
    residue_name_handling,
    line_info,
):

    axis_values = []
    for axis_field in list("LPWBEJU"):
        header = f"{axis}.{axis_field}"

        if header in heading_indices:
            field_index = heading_indices[header]
            value = tcl_list[field_index]
        else:
            value = None
        if axis_field == "L":
            axis_values.append(
                _read_atom_label(
                    value,
                    chain_code,
                    sequence_lookup,
                    molecule_type,
                    residue_name_type,
                    residue_name_handling,
                    line_info,
                )
            )
        elif axis_field == "P":
            axis_values.append(_read_shift(value, line_info))
        elif axis_field in "WJU":
            pass
        elif axis_field == "E":
            merit = value
            axis_values.append(merit)
    return axis_values


def _read_shift(value, line_info):
    if not is_float(value):
        msg = f"""
                    in file {line_info.file_name} at the line {line_info.line_no}
                    i expected a float for the peak position, but i got the value '{value}'
                    the line was
                    {line_info.line}
                """
        exit_error(msg)
    shift = float(value)
    return shift


def _read_atom_label(
    value,
    chain_code,
    sequence_lookup,
    molecule_type,
    residue_name_type,
    residue_handling,
    line_info,
):
    if isinstance(value, str):
        label = value
    else:
        label = value[0] if value else "?"
    label = label.strip()

    if "." in label and len(label.split(".")) == 2:
        residue_name_residue_number, atom_name = label.split(".")
        file_residue_name, sequence_code = strip_characters_left(
            residue_name_residue_number, string.ascii_letters
        )
    else:
        sequence_code = None
        atom_name = None
        file_residue_name = None if label == "?" else label

    _exit_if_sequence_code_isnt_int_or_none(line_info, sequence_code)

    if sequence_code:
        if is_int(sequence_code):
            sequence_code = int(sequence_code)

    if sequence_code:
        sequence_residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, sequence_lookup
        )
    else:
        sequence_residue_name = None

    if file_residue_name and sequence_code:
        file_residue_name = _translate_1_to_3_or_or_exit_bad(
            file_residue_name, molecule_type, residue_name_type, line_info
        )

    file_residue_name = _correct_residue_name_case_if_required(
        residue_handling, file_residue_name, sequence_residue_name
    )

    _warn_of_residue_name_mismatch_if_required(
        file_residue_name, sequence_residue_name, residue_handling, line_info
    )

    residue_handling = [
        option
        for option in residue_handling
        if option
        not in {ResidueNameHandlingOption.CASE, ResidueNameHandlingOption.WARN}
    ]

    residue_handling = residue_handling.pop()

    if file_residue_name:
        three_letter_file_residue_name = (
            translate_1_to_3(file_residue_name, molecule_type)[0]
            if len(file_residue_name) == 1
            else file_residue_name
        )

        if three_letter_file_residue_name and sequence_residue_name:
            if three_letter_file_residue_name.lower() == sequence_residue_name.lower():
                file_residue_name = three_letter_file_residue_name

    if (
        file_residue_name
        and sequence_residue_name
        and (sequence_residue_name != file_residue_name)
    ):
        if residue_handling == ResidueNameHandlingOption.STOP:
            _exit_residue_name_mismatch(
                file_residue_name, sequence_residue_name, line_info
            )
            residue_name = sequence_residue_name
        elif residue_handling == ResidueNameHandlingOption.PSEUDO:
            residue_name = file_residue_name
            sequence_code = f"@{sequence_code}"
        elif residue_handling == ResidueNameHandlingOption.UNASSIGN:
            residue_name = None
            sequence_code = None
            chain_code = None
            atom_name = None
        elif residue_handling == ResidueNameHandlingOption.SEQUENCE:
            residue_name = sequence_residue_name
        elif residue_handling == ResidueNameHandlingOption.CONTINUE:
            residue_name = file_residue_name
        else:
            raise Exception("unexpected")
    elif not file_residue_name and sequence_residue_name:
        residue_name = sequence_residue_name
    else:
        residue_name = file_residue_name

    chain_code = chain_code if sequence_code or atom_name else UNUSED
    sequence_code = UNUSED if not sequence_code else sequence_code
    atom_name = UNUSED if not atom_name else atom_name
    residue_name = UNUSED if not residue_name else residue_name

    atom = AtomLabel(
        SequenceResidue(chain_code, sequence_code, residue_name),
        atom_name,
    )

    return atom


def _warn_of_residue_name_mismatch_if_required(
    file_residue_name, sequence_residue_name, residue_handling, line_info
):
    if ResidueNameHandlingOption.WARN in residue_handling:
        msg = (
            f"WARNING: at line {line_info.line_no} in file {line_info.file_name} the residue name "
            + f"{file_residue_name} doesn't match the sequence residue name {sequence_residue_name}"
        )
        print(msg, file=sys.stderr)


def _correct_residue_name_case_if_required(
    residue_handling, file_residue_name, sequence_residue_name
):
    if ResidueNameHandlingOption.CASE in residue_handling:
        if (
            sequence_residue_name
            and file_residue_name
            and sequence_residue_name.lower() == file_residue_name.lower()
        ):
            file_residue_name = sequence_residue_name
    return file_residue_name


def _exit_residue_name_mismatch(file_residue_name, sequence_residue_name, line_info):
    msg = f"""
                    in file {line_info.file_name} at line {line_info.line_no}
                    the residue name {file_residue_name} doesn't match the one in the sequence {sequence_residue_name}
                    the line was
                    {line_info.line}
                """
    exit_error(msg)


def _exit_no_residue_name(line_info):
    msg = f"""
            in file {line_info.file_name} at line {line_info.line_no}
            residue name is not defined for the peak
            the line was
            {line_info.line}
        """
    exit_error(msg)


def _translate_1_to_3_or_or_exit_bad(
    residue_name, molecule_type, residue_name_type, line_info
):
    if residue_name:
        if residue_name_type == ResidueNameTypeOption.AUTO:
            if len(residue_name) > 1:
                residue_name = translate_1_to_3(residue_name, molecule_type)
        elif residue_name_type == ResidueNameTypeOption.SINGLE:
            if len(residue_name) > 1:
                msg = f"""
                    in file {line_info.file_name} at line {line_info.line_no}
                    the residue name {residue_name} is longer than one character
                    and you have chosen to use single letter residue names
                    the line was
                    {line_info.line}
                """
                exit_error(msg)
            residue_name = translate_1_to_3(residue_name, molecule_type)

    return residue_name


def _exit_if_sequence_code_isnt_int_or_none(line_info, residue_number):
    if residue_number and not is_int(residue_number):
        msg = f"""
            in file {line_info.file_name} at line {line_info.line_no}
            I expected an integer for the residue number nut got {residue_number}
            in line
            {line_info.line}
        """
        exit_error(msg)


def _exit_if_sequence_code_not_in_the_sequence(
    chain_code, residue_number, chain_start, chain_end, line_info
):
    if residue_number and (residue_number < chain_start or residue_number > chain_end):
        msg = f"""
            in the file {line_info.file_name} at {line_info.line_no}
            a residue type is not defined for chain: {chain_code} and residue number: {residue_number}
            the line was
            {line_info.line}
        """
        exit_error(msg)


def _read_values_for_peak(line, heading_indices):
    peak_values = {}
    for value_field in ["vol", "int", "stat", "comment", "flag0"]:
        if value_field in heading_indices:
            field_index = heading_indices[value_field]
            value = line[field_index]
        else:
            value = None

        if value_field == "vol":
            peak_values["volume"] = float(value)
        elif value_field == "int":
            peak_values["height"] = float(value)
        elif value_field == "stat":
            if value is None:
                peak_values["deleted"] = False
            else:
                peak_values["deleted"] = int(value) < 0
        elif value_field == "comment":
            comment = value[0].strip("'") if value else ""
            peak_values["comment"] = comment
        elif value_field == "flag0":
            pass

    return peak_values


def check_num_fields(fields, number, field_type, line, line_no):
    if len(fields) != number:
        msg = f"Expected {number} {field_type} got {len(fields)} for line: {line} at line {line_no}"
        exit_error(msg)


def _sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result


def _get_isotope_code_or_exit(axis, axis_codes):
    if axis >= len(axis_codes):
        msg = f"can't find isotope code for axis {axis + 1} got axis codes {','.join(axis_codes)}"
        exit_error(msg)
    axis_code = axis_codes[axis]
    return axis_code


def round_to_nearest_and_distance(number, nearest):
    rounded = round(number / nearest) * nearest
    return rounded, abs(rounded - number)


def _guess_spectrometer_frequency(peak_list):
    frequencies = [
        float(frequency)
        for frequency in peak_list.peak_list_data.spectrometer_frequencies
    ]
    max_frequency = max(frequencies)

    divisor = 10 if max_frequency < 240 else 50

    distance_and_frequency = {}
    for spectrometer_frequency in frequencies:
        rounded_spectrometer_frequency, distance = round_to_nearest_and_distance(
            spectrometer_frequency, divisor
        )
        distance_and_frequency[distance] = spectrometer_frequency

    min_distance = min(distance_and_frequency.keys())
    return distance_and_frequency[min_distance]


def _spectrometer_frequencies_to_axis_codes(spectrometer_frequency, peak_list):
    isotopes = []
    for frequency in peak_list.peak_list_data.spectrometer_frequencies:
        frequency = float(frequency)
        ratio = frequency / spectrometer_frequency
        ratio_distances = {
            isotope: abs(ratio - gamma_ratio)
            for isotope, gamma_ratio in GAMMA_RATIOS.items()
        }
        closest_isotope = min(ratio_distances, key=ratio_distances.get)
        isotopes.append(closest_isotope)

    return isotopes


# TODO this should use peak_lib peaks_to_frame instead
def _create_spectrum_frame(entry_name, peak_list, chain_code):

    spectrometer_frequency = _guess_spectrometer_frequency(peak_list)
    peak_bounds = _get_peak_bounds(peak_list)
    axis_isotopes = _spectrometer_frequencies_to_axis_codes(
        spectrometer_frequency, peak_list
    )

    category = "nef_nmr_spectrum"
    frame_code = f"{category}_{entry_name}"
    frame = Saveframe.from_scratch(frame_code, category)

    frame.add_tag("sf_category", category)
    frame.add_tag("sf_framecode", frame_code)
    frame.add_tag("num_dimensions", peak_list.peak_list_data.num_axis)
    frame.add_tag("chemical_shift_list", constants.NEF_UNKNOWN)
    loop = Loop.from_scratch("nef_spectrum_dimension")
    frame.add_loop(loop)
    list_tags = (
        "dimension_id",
        "axis_unit",
        "axis_code",
        "spectrometer_frequency",
        "spectral_width",
        "value_first_point",
        "folding",
        "absolute_peak_positions",
        "is_acquisition",
    )
    loop.add_tag(list_tags)
    list_data = peak_list.peak_list_data

    expanded_peak_bounds = {
        axis: peak_bounds.expand(10.0) for axis, peak_bounds in peak_bounds.items()
    }
    expanded_peak_bounds = {
        axis: peak_bounds.ensure_width(1.0)
        for axis, peak_bounds in expanded_peak_bounds.items()
    }
    for i in range(list_data.num_axis):
        row = {
            "dimension_id": i + 1,
            "axis_unit": "ppm",
            "axis_code": axis_isotopes[i],
            "spectrometer_frequency": list_data.spectrometer_frequencies[i],
            "spectral_width": round(expanded_peak_bounds[i].width, 5),
            "value_first_point": round(expanded_peak_bounds[i].max, 5),
            "folding": "circular",
            "absolute_peak_positions": "true",
            "is_acquisition": NEF_UNKNOWN,
        }
        loop.add_data(
            [
                row,
            ]
        )

    loop = Loop.from_scratch("nef_spectrum_dimension_transfer")
    frame.add_loop(loop)
    transfer_dim_tags = ("dimension_1", "dimension_2", "transfer_type")
    loop.add_tag(transfer_dim_tags)
    loop = Loop.from_scratch("nef_peak")
    frame.add_loop(loop)

    # TODO: put this in a sane order!
    peak_tags = [
        "index",
        "peak_id",
        "volume",
        "volume_uncertainty",
        "height",
        "height_uncertainty",
    ]
    position_tags = [
        (f"position_{i + 1}", f"position_uncertainty_{i + 1}")
        for i in range(list_data.num_axis)
    ]
    position_tags = itertools.chain(*position_tags)
    atom_name_tags = [
        (
            f"chain_code_{i + 1}",
            f"sequence_code_{i + 1}",
            f"residue_name_{i + 1}",
            f"atom_name_{i + 1}",
        )
        for i in range(list_data.num_axis)
    ]
    atom_name_tags = itertools.chain(*atom_name_tags)
    tags = [*peak_tags, *position_tags, *atom_name_tags]

    loop.add_tag(tags)
    for i, peak in enumerate(peak_list.peaks):
        peak_values = peak["values"]
        if peak_values.deleted:
            continue

        positions = {}
        for tag in tags:
            if tag.split("_")[0] == "position" and len(tag.split("_")) == 2:
                index = int(tag.split("_")[-1]) - 1
                positions[tag] = peak[index].ppm

        chain_codes = {}
        for tag in tags:
            if tag.split("_")[:2] == ["chain", "code"]:
                index = int(tag.split("_")[-1]) - 1
                residue = peak[index].atom_labels.residue
                chain_code = residue.chain_code
                chain_code = chain_code if chain_code is not None else chain_code
                chain_code = chain_code if chain_code else "."
                chain_code_prefix = residue.chain_code_prefix
                chain_code = f"{chain_code_prefix}{chain_code}"

                chain_codes[tag] = chain_code

        sequence_codes = {}
        for tag in tags:
            if tag.split("_")[:2] == ["sequence", "code"]:
                index = int(tag.split("_")[-1]) - 1
                residue = peak[index].atom_labels.residue
                sequence_code = residue.sequence_code
                sequence_code = sequence_code if sequence_code else "."
                sequence_code_prefix = residue.sequence_code_prefix
                sequence_code = f"{sequence_code_prefix}{sequence_code}"
                sequence_codes[tag] = sequence_code

        residue_names = {}
        for tag in tags:
            if tag.split("_")[:2] == ["residue", "name"]:
                index = int(tag.split("_")[-1]) - 1

                # TODO: there could be more than 1 atom label here and this should be a list...
                residue_name = peak[index].atom_labels.residue.residue_name
                residue_name = residue_name if residue_name else "."
                residue_names[tag] = residue_name

        atom_names = {}
        for tag in tags:
            if tag.split("_")[:2] == ["atom", "name"]:
                index = int(tag.split("_")[-1]) - 1

                atom_name = peak[index].atom_labels.atom_name
                atom_name = atom_name if atom_name else "."
                atom_names[tag] = atom_name

        row_dict = {
            "index": i + 1,
            "peak_id": peak_values.serial + 1,
            **chain_codes,
            **sequence_codes,
            **residue_names,
            **atom_names,
            **positions,
            "volume": peak_values.volume,
            "volume_uncertainty": NEF_UNKNOWN,
            "height": peak_values.height,
            "height_uncertainty": NEF_UNKNOWN,
        }
        loop.add_data(
            [
                row_dict,
            ]
        )
    return frame


# TODO: can be replaced by str.removesuffix when min python version >= 3.9
def _remove_suffix(string: str, suffix: str) -> str:

    result = string
    if string.endswith(suffix):
        result = string[: -len(suffix)]

    return result


def _make_peak_list_frame_name(peaks_list, file_name, frame_name_source):
    if frame_name_source == FrameNameOption.SPECTRUM:
        frame_name = peaks_list.peak_list_data.data_set.replace(" ", "_")
        frame_name = _remove_suffix(frame_name, ".nv")
        frame_name = frame_name.replace(".", "_")
    else:
        file_name = Path(file_name).stem
        frame_name = file_name.replace(" ", "_")
    return frame_name


def _fixup_residue_handling_or_exit_bad(residue_handling):

    number_residue_handling_options = len(residue_handling)
    if number_residue_handling_options == 0:
        residue_handling = [
            ResidueNameHandlingOption.STOP,
        ]

    active_residue_options = _remove_residue_handling_modifiers(residue_handling)
    if len(active_residue_options) != 1:
        _exit_too_many_residue_handling_options(residue_handling)
    return residue_handling


def _fixup_residue_number_handling_or_exit_bad(residue_number_handling):

    number_residue_handling_options = len(residue_number_handling)
    if number_residue_handling_options == 0:
        residue_number_handling = [
            ResidueNameHandlingOption.PSEUDO,
            ResidueNumberHandlingOption.GLOBAL_WARN,
            ResidueNumberHandlingOption.COMMENT,
        ]

    active_residue_options = _remove_residue_number_handling_modifiers(
        residue_number_handling
    )
    if len(active_residue_options) > 2:
        _exit_too_many_residue_handling_options(residue_number_handling)
    return residue_number_handling


def _remove_residue_number_handling_modifiers(residue_number_handling):
    return [
        option
        for option in residue_number_handling
        if option
        not in {
            ResidueNumberHandlingOption.WARN,
            ResidueNumberHandlingOption.GLOBAL_WARN,
        }
    ]


def _remove_residue_handling_modifiers(residue_handling):
    return [
        option
        for option in residue_handling
        if option
        not in {ResidueNameHandlingOption.CASE, ResidueNameHandlingOption.WARN}
    ]


def _exit_too_many_residue_handling_options(residue_handling):
    active_residue_handling_options = _remove_residue_handling_modifiers(
        residue_handling
    )
    number_active_residue_handling_options = len(active_residue_handling_options)
    if number_active_residue_handling_options > 1:
        msg = f"""
            you have specified residue handling options  {','.join(active_residue_handling_options)}
            that are not compatible with each other, you can only combine options with case and warn...
            """
    exit_error(msg)


@dataclass
class AxisBounds:
    min: float
    max: float

    @property
    def middle(self):
        return (self.max + self.min) / 2.0

    @property
    def width(self):
        return self.max - self.min

    def expand(self, percentage):
        width = self.max - self.min
        extra = width * (percentage / 100.0)

        return AxisBounds(self.min - extra, self.max + extra)

    def ensure_width(self, required_width):
        new_min = self.min
        new_max = self.max
        if self.width < required_width:
            extra_width = required_width - self.width
            extra_width_2 = extra_width / 2.0
            new_min = self.min - extra_width_2
            new_max = self.max + extra_width_2

        return AxisBounds(new_min, new_max)


def _get_peak_bounds(peak_list):
    shifts_by_axis = {}
    axes = OrderedSet()
    result = {}
    for peak in peak_list.peaks:

        for item, peak in peak.items():
            if is_int(item):
                shifts_by_axis.setdefault(item, []).append(peak.ppm)
                axes.add(item)

    for axis in axes:
        axis_min = min(shifts_by_axis[axis])
        axis_max = max(shifts_by_axis[axis])
        result[axis] = AxisBounds(axis_min, axis_max)

    return result



#!/usr/bin/env python3
# Importing libraries
import argparse
import fcntl
import os
import sys
from fnmatch import fnmatch

from pynmrstar import Entry

CHAIN_CODE = "chain_code"

SEQUENCE_CODE = "sequence_code"

parser = None

EXIT_ERROR = 1


def exit_error(msg):
    if parser:
        parser.print_help()
    print()
    print(f"ERROR: {msg}")
    print("exiting...")
    sys.exit(EXIT_ERROR)


target_frames = {
    "nef_chemical_shift_list": ["_nef_chemical_shift"],
    "nef_nmr_spectrum": ["_nef_peak"],
}


def offset_chemical_shifts(entry, args):

    if not args.target:
        exit_error(
            "WARNING: i need a target nucleus or atom name provided by -t / --target,  such as *CB* or 13C"
        )

    target_loops = []
    for frame_data in entry.frame_dict.values():
        if (frame_category := frame_data.category) in target_frames:
            for loop in frame_data.loops:
                if (loop_category := loop.category) in target_frames[frame_category]:
                    target_loops.append(((frame_category, loop_category), loop))

    for loop_type, target_loop in target_loops:
        if loop_type == ("nef_chemical_shift_list", "_nef_chemical_shift"):
            chain_index = target_loop.tag_index("chain_code")
            value_index = target_loop.tag_index("value")
            element_index = target_loop.tag_index("element")
            isotope_index = target_loop.tag_index("isotope_number")
            atom_name_index = target_loop.tag_index("atom_name")

            for line in target_loop:
                isotope = line[isotope_index] + line[element_index]
                chain = line[chain_index]
                value = line[value_index]
                atom_name = line[atom_name_index]

                if chain == args.chain_code:
                    if args.target == isotope or fnmatch(atom_name, args.target):
                        value = float(value) + args.offset
                        line[value_index] = value

        if loop_type == ("nef_nmr_spectrum", "_nef_peak"):
            exit_error("Error: not implemented for peak lists yet")

        # for loop_data in frame_data.loop_dict.values():
        #     print(loop_data.category)
        # if SEQUENCE_CODE in loop_data.tags:
        #     sequence_index = loop_data.tag_index(SEQUENCE_CODE)
        #     chain_index = loop_data.tag_index(CHAIN_CODE)
        #     for line in loop_data:
        #         if line[chain_index] == chain:
        #             sequence = line[sequence_index]
        #             if  '@' not in sequence:
        #                 try:
        #                     sequence = int(sequence)
        #                 except ValueError:
        #                     pass
        #             if isinstance(sequence, int):
        #                 sequence = sequence + offset
        #                 line[sequence_index] = str(sequence)


def read_args():
    global parser
    parser = argparse.ArgumentParser(description="Offset chemical shifts")
    parser.add_argument(
        "-c",
        "--chain",
        metavar="CHAIN",
        type=str,
        default="A",
        dest="chain",
        help="chain in which to offset shifts",
    )
    parser.add_argument(
        "-o",
        "--offset",
        metavar="OFFSET",
        type=float,
        default=0.0,
        dest="offset",
        help="offset to add to shifts",
    )
    parser.add_argument(
        "-t",
        "--target",
        metavar="TARGET",
        type=str,
        default=None,
        dest="target",
        help="nucleus or atom to offset",
    )
    parser.add_argument(
        "-s",
        "--spectra",
        metavar="OFFSET",
        type=str,
        default="*",
        dest="spectra",
        action="append",
        nargs="+",
        help="nucleus  or atom to offset",
    )
    parser.add_argument(metavar="FILES", nargs=argparse.REMAINDER, dest="files")

    return parser.parse_args()


def check_stream():
    # make stdin a non-blocking file
    fd = sys.stdin.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

    try:
        result = sys.stdin.read()
    except Exception:
        result = ""

    return result


if __name__ == "__main__":

    args = read_args()

    print(args)

    is_tty = sys.stdin.isatty()

    if is_tty and len(args.files) == 0:
        parser.print_help()
        print()
        msg = """I require at least 1 argument or input stream with a chemical_shift_list frame"""

        exit_error(msg)

    entries = []
    try:
        if len(args.files) == 0:
            lines = check_stream()
            if len(lines) != 0:
                entries.append(Entry.from_string(lines))
            else:
                exit_error("Error: input appears to be empty")
        else:
            for file in args.files:
                entries.append(Entry.from_file(file))
    except OSError as e:
        msg = f"couldn't open target nef file because {e}"
        exit_error(msg)

    for entry in entries:
        offset_chemical_shifts(entry, args)

        print(entry)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(
        app, name="modelfree", help="- write modelfree [relaxation data]"
    )

    app.add_typer(
        export_app, name="export", help="- export modelfree [relaxation data]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.modelfree.exporters.data  # noqa: F401



import typer
from pynmrstar.loop import Loop

from nef_pipelines.lib.test_lib import isolate_loop, path_in_test_data, run_and_report
from nef_pipelines.tools.frames.filter import filter

app = typer.Typer()
app.command()(filter)


def test_filter_unassigned():
    path = path_in_test_data(__file__, "assigned_partially_and_unassigned.nef")
    result = run_and_report(app, ["--in", path])

    loop = isolate_loop(
        result.stdout, "nef_rdc_restraint_list_test_1", "nef_rdc_restraint"
    )

    loop = Loop.from_string(loop)
    assert len(loop) == 1
    assert "1  1  .  A  1  ALA  H  A  1   ALA  N  1  -4.2  0.22".split() in loop.data


def test_filter_unassigned_partial():
    path = path_in_test_data(__file__, "assigned_partially_and_unassigned.nef")
    result = run_and_report(app, f"--in {path} --state partial".split())

    loop = isolate_loop(
        result.stdout, "nef_rdc_restraint_list_test_1", "nef_rdc_restraint"
    )

    loop = Loop.from_string(loop)
    assert len(loop) == 2
    assert (
        "1   1   .   A   1   ALA   H   A   1   ALA   N   1   -4.2   0.22".split()
        in loop.data
    )
    assert (
        "2   2   .   .   .   .     .   A   2   TRP   N   1   -5.2   0.33".split()
        in loop.data
    )


def test_filter_assigned():
    path = path_in_test_data(__file__, "assigned_partially_and_unassigned.nef")
    result = run_and_report(app, ["--in", path, "--assigned"])

    loop = isolate_loop(
        result.stdout, "nef_rdc_restraint_list_test_1", "nef_rdc_restraint"
    )

    loop = Loop.from_string(loop)
    assert len(loop) == 1
    assert "3  3  .  .  .  .    .  .  .   .    .  1   3.1  0.4".split() in loop.data


def test_filter_assigned_full():
    path = path_in_test_data(__file__, "assigned_partially_and_unassigned.nef")
    result = run_and_report(app, f"--in {path} --assigned --state full".split())

    loop = isolate_loop(
        result.stdout, "nef_rdc_restraint_list_test_1", "nef_rdc_restraint"
    )

    loop = Loop.from_string(loop)
    assert len(loop) == 2
    assert " 2  2  .  .  .  .    .  A  2   TRP  N  1  -5.2  0.33".split() in loop.data
    assert "3  3  .  .  .  .    .  .  .   .    .  1   3.1  0.4".split() in loop.data



import platform
import sys
from datetime import date
from textwrap import dedent, indent

import typer

from nef_pipelines.lib.util import get_version
from nef_pipelines.tools.help import help_app

VERBOSE_HELP = """\
    display more verbose information about NEF-Pipelines and its environment,
    [ignored with the --version option]"
"""


# noinspection PyUnusedLocal
@help_app.command()
def about(
    version: bool = typer.Option(
        False, "--version", help="only display the current version"
    ),
    verbose: bool = typer.Option(False, "-v", "--verbose", help=VERBOSE_HELP),
):
    "- info about this NEF-Pipelines installation"

    print(cmd(version, verbose))


def cmd(version, verbose=False):
    version_data = get_version()

    if version:
        result = str(version_data)
    else:
        NEF_PIPLINES_URL = "https://github.com/varioustoxins/NEF-Pipelines"

        if verbose:
            # TODO add the modules in requirements.txt and their versions
            extra_msg = f"""\

                Environment
                -----------

                Python version:   {platform.python_version()}
                Python path:      {sys.executable}
                Operating system: {platform.platform(aliased=True, terse=True)}

            """
        else:
            extra_msg = ""
        extra_msg = dedent(extra_msg)

        msg = f"""\

            This is NEF-Pipelines version {version_data}

            For information about this program and the code please see: {NEF_PIPLINES_URL}
            For bug reporting please goto: {NEF_PIPLINES_URL}/issues

            If you like this project please star it on the projects github page (listed above)!
            If you want to cite the project please cite it as

            G. S. Thompson {NEF_PIPLINES_URL} ({date.today().year})

            regards Gary (aka varioustoxins)
        """

        msg = dedent(msg)

        max_width = max([len(line) for line in msg.split("\n")])

        header = "=" * (max_width + 4)

        result = "\n".join([header, indent(msg, "  "), indent(extra_msg, "  "), header])

    return result



from collections import Counter
from dataclasses import dataclass
from itertools import chain, cycle, islice
from pathlib import Path
from typing import Iterable, List

import typer
from fastaparser import Reader
from ordered_set import OrderedSet
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    BadResidue,
    MoleculeType,
    get_chain_code_iter,
    offset_chain_residues,
    sequence_3let_to_res,
    sequence_to_nef_frame,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import (
    NEWLINE,
    STDIN,
    exit_error,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.fasta import import_app

CHAIN_STARTS_HELP = """first residue number of sequences can be a comma separated list or added multiple times
                       note the offsets are applied on top of any chain starts present in the file headers unless
                       the option --no-header is used"""

NO_HEADER_HELP = """don't read the NEF-Pipelines sequence header if present
                    the format of the header is >{entry_name} CHAIN: {chain} | START RESIDUE: {chain_start}
                    an example would be >test CHAIN: A | START: -2 for the entry test where chain A starts at -2
                   """

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# todo add comment to other sequences etc
@import_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="chain codes to use for the imported chains, can be a a comma separated list or can be called "
        "multiple times",
        metavar="<CHAIN-CODES>",
    ),
    starts: List[str] = typer.Option(
        [],
        "--starts",
        help=CHAIN_STARTS_HELP,
        metavar="<START>",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    # TODO: unused inputs!
    entry_name: str = typer.Option(None, help="a name for the entry if required"),
    in_file: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="input to read NEF data from [- is stdin]",
    ),
    molecule_types: List[MoleculeType] = typer.Option(
        [
            MoleculeType.PROTEIN,
        ],
        "--molecule-type",
        help="molecule type one of protein, dna, rna or  carbohydrate",
    ),
    no_header: bool = typer.Option(False, "--no-header", help=NO_HEADER_HELP),
    file_names: List[Path] = typer.Argument(
        ..., help="the file to read", metavar="<FASTA-FILE>"
    ),
):
    """- convert fasta sequence to nef"""

    file_names = parse_comma_separated_options(file_names)

    chain_codes = parse_comma_separated_options(chain_codes)
    if not chain_codes:
        chain_codes = [
            "A",
        ]

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [
            False,
        ]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [
            False,
        ]

    starts = [int(elem) for elem in parse_comma_separated_options(starts)]
    if not starts:
        starts = [
            1,
        ]

    molecule_types = parse_comma_separated_options(molecule_types)
    if not molecule_types:
        molecule_types = [
            MoleculeType.PROTEIN,
        ]

    entry = read_or_create_entry_exit_error_on_bad_file(in_file)

    entry = pipe(
        entry,
        chain_codes,
        starts,
        no_chain_starts,
        no_chain_ends,
        molecule_types,
        no_header,
        file_names,
        entry_name,
    )

    print(entry)


def pipe(
    entry: Entry,
    chain_codes: List[str],
    starts: List[int],
    no_chain_starts: List[bool],
    no_chain_ends: List[bool],
    molecule_types: List[MoleculeType],
    no_header: bool,
    file_names,
    entry_name: str,
):
    fasta_frames = []

    fasta_residues, read_entry_name = _read_sequences(
        file_names, chain_codes, molecule_types, not no_header
    )

    if read_entry_name and not entry_name:
        entry_name = "__".join(read_entry_name)
        if not entry_name:
            entry_name = "fasta"
        entry.entry_id = entry_name
    elif entry_name:
        entry.entry_id = entry_name
    else:
        entry.entry_id = "fasta"

    read_chain_codes = residues_to_chain_codes(fasta_residues)

    offsets = _get_sequence_offsets(read_chain_codes, starts)

    fasta_residues = offset_chain_residues(fasta_residues, offsets)

    fasta_residues = sorted(fasta_residues)

    fasta_frames.append(
        sequence_to_nef_frame(fasta_residues, no_chain_starts, no_chain_ends)
    )

    entry = add_frames_to_entry(entry, fasta_frames)

    return entry


def _get_sequence_offsets(chain_codes: List[str], starts: List[int]):

    offsets = [start - 1 for start in starts]
    cycle_starts = chain(
        offsets,
        cycle(
            [
                0,
            ]
        ),
    )
    offsets = list(islice(cycle_starts, len(chain_codes)))

    return {chain_code: offset for chain_code, offset in zip(chain_codes, offsets)}


def residues_to_chain_codes(residues: List[SequenceResidue]) -> List[str]:
    return list(OrderedSet([residue.chain_code for residue in residues]))


@dataclass
class Sequence:
    entry_id: str
    comment: str
    sequence: List[str]
    chain_code: str = None
    starts: int = 1


# could do with taking a list of offsets
# noinspection PyUnusedLocal
#
def _parse_fasta(fasta_sequence, parse_header: bool = True) -> Sequence:
    definition = fasta_sequence.formatted_definition_line()
    bar_count = definition.count("|")
    last_part = definition.split()[-1]
    is_bracketed = last_part.startswith("(") and last_part.endswith(")")
    is_last_field_number = last_part.strip("()").isdigit()

    chain_code = None
    start = 1

    if fasta_sequence.description.startswith("NEFPLS") and parse_header:
        fields = fasta_sequence.description.split("|")
        for field in fields:

            field = field.strip()

            if field.startswith("CHAIN:"):
                chain_code = field.split()[-1]

            if field.startswith("START:"):
                if is_int(field[-1]):
                    start = int(field.split()[-1])

        entry_id, comment = fasta_sequence.id, fasta_sequence.description

    elif bar_count == 3 and is_bracketed and is_last_field_number:
        fields = definition.split("|")
        entry_id = fields[0].lstrip(">")
        comment = "__".join(fields[1:])
    else:
        sequence_id = "".join([c if c.isalnum() else "_" for c in fasta_sequence.id])
        entry_id, comment = sequence_id, fasta_sequence.description

    letters = [letter_code.letter_code for letter_code in fasta_sequence.sequence]
    return Sequence(entry_id, comment, letters, chain_code, start)


@dataclass
class SequenceInfo:
    entry_id: str
    chain_code: str
    start: int


def _exit_if_there_are_file_and_input_chain_codes(
    file_chain_codes, input_chain_codes, file_paths
):
    input_chain_codes = [
        chain_code for chain_code in input_chain_codes if chain_code is not None
    ]
    file_chain_codes = [
        chain_code for chain_code in file_chain_codes if chain_code is not None
    ]
    if file_chain_codes and input_chain_codes:
        msg = f"""
            there are chain codes in the files and input arguments
            the chain codes in the files are {','.join(file_chain_codes)}
            the chain codes in the input arguments are {','.join(input_chain_codes)}
            the files are
            {NEWLINE.join([str(file_path) for file_path in file_paths])}
            chain codes must come from files or inputs not both!
            """
        exit_error(msg)


def _exit_if_there_are_gaps_in_file_chain_codes(file_chain_codes, file_paths):

    file_chain_codes = [
        chain_code for chain_code in file_chain_codes if chain_code is not None
    ]
    if len(file_chain_codes) != len(file_chain_codes):
        msg = f"""
            there are sequences without chain codes in the files
            {NEWLINE.join([str(file_path) for file_path in file_paths])}
            either all sequences in files must have a chain code or none
            """
        exit_error(msg)


def _read_sequences(
    file_paths: List[Path],
    chain_codes: Iterable[str],
    molecule_types: List[MoleculeType],
    parse_header=False,
) -> List[SequenceResidue]:

    sequence_records = []

    for file_path in file_paths:
        try:
            with open(file_path) as handle:
                try:
                    reader = Reader(handle)
                    for fasta_sequence in reader:
                        sequence_records.append(
                            _parse_fasta(fasta_sequence, parse_header)
                        )

                except Exception as e:
                    # check if relative to os.getcwd
                    exit_error(f"Error reading fasta file {str(file_path)}", e)
        except IOError as e:
            exit_error(f"couldn't open {file_path} because:\n{e}", e)

    number_sequences = len(sequence_records)
    number_molecule_types = len(molecule_types)

    if number_molecule_types == 1:
        molecule_types = molecule_types * number_sequences
    elif number_molecule_types == 0 or number_molecule_types < number_sequences:
        msg = f"""
            number molecule types [{number_molecule_types}] is different from number of chains {number_sequences}
        """
        exit_error(msg)

    residues = OrderedSet()
    # read as many chain codes as there are sequences
    # https://stackoverflow.com/questions/16188270/get-a-fixed-number-of-items-from-a-generator

    file_chain_codes = [
        sequence_record.chain_code for sequence_record in sequence_records
    ]

    _exit_if_there_are_gaps_in_file_chain_codes(file_chain_codes, file_paths)

    # this keeps the default input chain code as A
    file_chain_codes = [
        chain_code for chain_code in file_chain_codes if chain_code is not None
    ]
    if file_chain_codes and chain_codes == ["A"]:
        chain_codes = []

    _exit_if_there_are_file_and_input_chain_codes(
        file_chain_codes, chain_codes, file_paths
    )

    _exit_if_there_are_replicate_chain_codes_from_files(
        file_chain_codes, chain_codes, file_paths
    )

    chain_codes = file_chain_codes if file_chain_codes else chain_codes

    chain_code_iter = get_chain_code_iter(chain_codes)
    for sequence_record, chain_code, molecule_type in zip(
        sequence_records, chain_code_iter, molecule_types
    ):

        try:
            sequence_3_let = translate_1_to_3(
                sequence_record.sequence, molecule_type=molecule_type
            )
        except BadResidue as e:
            exit_error(
                f"Error translating sequence {sequence_record} to 3 letter code {e}",
                e,
            )

        chain_code = (
            sequence_record.chain_code if sequence_record.chain_code else chain_code
        )
        chain_residues = sequence_3let_to_res(sequence_3_let, chain_code)

        chain_residues = offset_chain_residues(
            chain_residues,
            {sequence_record.chain_code: sequence_record.starts - 1},
        )

        residues.update(chain_residues)

    entry_names = [
        sequence_record.entry_id
        for sequence_record in sequence_records
        if sequence_record.entry_id
    ]

    return residues, entry_names


def _exit_if_there_are_replicate_chain_codes_from_files(
    file_chain_codes, input_chain_codes, file_paths
):

    file_chain_counts = Counter(file_chain_codes)
    file_chain_replicates = {
        chain_code: count
        for chain_code, count in file_chain_counts.items()
        if count > 1 and chain_code is not None
    }
    input_chain_counts = Counter(input_chain_codes)
    input_chain_replicates = {
        chain_code: count
        for chain_code, count in input_chain_counts.items()
        if count > 1
    }

    if file_chain_replicates:
        repeats = " ".join(
            [
                f"{chain_code} [{count}]"
                for chain_code, count in file_chain_replicates.items()
            ]
        )

        msg = f"""
                some of the input chain codes are repeated: {repeats}
                {NEWLINE.join([str(file_path) for file_path in file_paths])}
                all chain codes must be unique
            """
        exit_error(msg)

    if input_chain_replicates:
        msg = f"""
            some of the chain codes: {','.join(input_chain_replicates.keys())}
            are repeated in the input each chain must have a unique chain code
        """

        exit_error(msg)

    all_chain_counts = Counter(file_chain_codes + input_chain_codes)
    all_chain_replicates = {
        chain_code: count
        for chain_code, count in all_chain_counts.items()
        if count > 1 and chain_code is not None
    }
    if all_chain_replicates:
        msg = f"""
                some of the chain codes are repeated: {','.join(all_chain_replicates.keys())}
                the chain codes come from a combination of the input arguments and the files
                {NEWLINE.join([str(file_paths) for file_path in file_paths])}
                all chain_codes_must_be_unique
            """
        exit_error(msg)



import typer

from nef_pipelines import nef_app

loops_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        loops_app,
        name="loops",
        help="- carry out operations on loops in nef frames",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.loops.trim  # noqa: F401



from collections import Counter
from enum import auto
from pathlib import Path
from typing import List

import typer
from pynmrstar import Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib.nef_frames_lib import NEF_PIPELINES_NAMESPACE
from nef_pipelines.lib.nef_lib import (
    create_nef_save_frame,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.structures import (
    UNUSED,
    AtomLabel,
    RdcRestraint,
    RdcTensorFrameData,
    SequenceResidue,
)
from nef_pipelines.lib.util import STDIN, exit_error
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    VALUES,
    get_column_indices,
    read_db_file_records,
    select_data_records,
    select_records,
)
from nef_pipelines.transcoders.pales import import_app

FLOAT_NAN = float("nan")
UNDEFINED_VECTOR_3D = [FLOAT_NAN, FLOAT_NAN, FLOAT_NAN]

_PALES_ANALYSIS_MODES = ("DC", "DC_CP", "SAUPE_CP")
_PALES_SIMULATION_MODES = ("STERIC", "STERIC_FREE")
_PALES_EXPECTED_REMARKS = (
    "REMARK Dipolar couplings.",
    "REMARK Molecular Alignment Simulation.",
    "REMARK Order matrix.",
    "REMARK Simulation parameters.",
)

_CHAINS_HELP = """\
    chains to import to, to add multiple chains use repeated calls, values can be comma separated there should
    be one chain per file if a single chain is provides for multiple files it is applied to all files
"""

PALES_EIGENVECTOR_AXIS_LIST = "XAXIS YAXIS ZAXIS".split()

app = typer.Typer()


_NEF_RDC_RESTRAINT_FRAME_CATEGORY = "nef_rdc_restraint_list"
_NEF_RDC_RESTRAINT_CATEGORY = "nef_rdc_restraint"
_NEFL_RDC_TENSOR_FRAME_FRAME_CATEGORY = NEF_PIPELINES_NAMESPACE + "_rdc_tensor"

_PALES_RDCS_EXPECTED_FIELDS = (
    "RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J DI D".split()
)


class RestraintOrigin(LowercaseStrEnum):
    MEASURED = auto()
    PREDICTED = auto()
    BACK_CALCULATED = auto()


class PalesDataType(LowercaseStrEnum):
    OBSERVED = auto()
    PREDICTED = auto()
    BACK_CALCULATED = auto()


seen_frame_names = Counter()


# noinspection PyUnusedLocal
@import_app.command()
def rdcs(
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        help="input to read NEF data from [- is stdin]",
    ),
    entry_name: str = typer.Option(
        "pales", "-e", "--entry-name", help="a name for the entry"
    ),
    chain_codes: List[str] = typer.Option(
        [],
        "-c",
        "--chains",
        help=_CHAINS_HELP,
        metavar="<CHAIN-CODE>",
    ),
    frame_name_template: str = typer.Option(
        "{file_name}_{data_type}",
        "-t",
        "--frame-name-template",
        help="name of the frame to add the restraints to, by defauylt templated from the file name",
    ),
    pales_input_files: List[Path] = typer.Argument(
        None,
        help="list of files to read PALES data from [- is stdin]",
    ),
):
    chain_codes = _normalise_chain_codes(chain_codes, pales_input_files)

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name=entry_name)

    entry = pipe(entry, pales_input_files, chain_codes, frame_name_template)

    print(entry)


def pipe(entry, pales_input_files, chain_codes, frame_name_template):

    for file_path, chain_code in zip(pales_input_files, chain_codes):

        db_text = _read_file_or_exit_error(file_path)

        db_records = read_db_file_records(db_text, file_path)
        rdc_data_measured = _parse_rdcs(
            db_records, (chain_code, chain_code), source=PalesDataType.OBSERVED
        )
        rdc_data_calculated = _parse_rdcs(
            db_records, (chain_code, chain_code), source=PalesDataType.PREDICTED
        )

        tensor_data = _parse_rdc_tensor(db_records)
        data_origin = _parse_data_origin(db_records)
        simulation_data = _parse_pales_simulation_data(db_records)

        calculated_frame_name, measured_frame_name = _make_frame_names(
            data_origin, file_path, frame_name_template
        )

        if data_origin != RestraintOrigin.PREDICTED:
            nef_measured_rdc_frame = _create_nef_output_frame(
                measured_frame_name, rdc_data_measured, "measured", None, None
            )
            entry.add_saveframe(nef_measured_rdc_frame)

        nef_calculated_rdc_frame = _create_nef_output_frame(
            calculated_frame_name,
            rdc_data_calculated,
            data_origin,
            simulation_data,
            tensor_data,
        )
        entry.add_saveframe(nef_calculated_rdc_frame)

    return entry


def _parse_pales_simulation_data(db_records):
    simulation_data = {}
    for db_record in select_data_records(db_records, "PALES"):
        simulation_data[db_record.values[1]] = " ".join(db_record.values[2:])
    return simulation_data


def _add_data_origin_to_frame(nef_rdc_frame, data_origin):
    nef_rdc_frame.add_tag("restraint_origin", data_origin)
    return nef_rdc_frame


def _add_da_and_dr_to_frame(nef_rdc_frame, tensor_data):
    if tensor_data.is_defined():
        nef_rdc_frame.add_tag("magnitude", tensor_data.da)
        nef_rdc_frame.add_tag("rhombicity", tensor_data.dr)
    return nef_rdc_frame


def _add_simulation_data_to_frame(nef_rdc_frame, simulation_data):
    if simulation_data:
        simulation_data_string = "\n".join(
            [f"{key}: {value}" for key, value in simulation_data.items()]
        )
        simulation_data_string = f"simulation data:\n\n{simulation_data_string}"

        nef_rdc_frame.add_tag("nefl_comment", simulation_data_string)
    return nef_rdc_frame


def _make_frame_names(data_origin, file_path, frame_name_template):
    if data_origin == RestraintOrigin.MEASURED:
        calculated_data_type = PalesDataType.BACK_CALCULATED
    else:
        calculated_data_type = PalesDataType.PREDICTED
    keys = {"file_name": file_path.stem, "data_type": calculated_data_type}
    calculated_frame_name = frame_name_template.format(**keys)
    keys = {"file_name": file_path.stem, "data_type": PalesDataType.OBSERVED}
    measured_frame_name = frame_name_template.format(**keys)
    if calculated_frame_name in seen_frame_names:
        calculated_frame_name = (
            f"{calculated_frame_name}__{seen_frame_names[calculated_frame_name]}"
        )
    seen_frame_names[calculated_frame_name] += 1
    if measured_frame_name in seen_frame_names:
        measured_frame_name = (
            f"{measured_frame_name}__{seen_frame_names[measured_frame_name]}"
        )
    seen_frame_names[measured_frame_name] += 1
    return calculated_frame_name, measured_frame_name


def _create_nef_output_frame(
    frame_name, rdc_data_measured, data_origin, simulation_data, tensor_data
):
    nef_rdc_frame = _RDCS_to_nef_saveframe(rdc_data_measured, frame_name)
    if tensor_data:
        nef_rdc_frame = _add_da_and_dr_to_frame(nef_rdc_frame, tensor_data)
        if tensor_data.is_defined():
            nef_tensor_loop = rdc_tensors_to_nef_loop(tensor_data)
            nef_rdc_frame.add_loop(nef_tensor_loop)

    nef_rdc_frame = _add_data_origin_to_frame(nef_rdc_frame, data_origin)
    if simulation_data:
        nef_rdc_frame = _add_simulation_data_to_frame(nef_rdc_frame, simulation_data)

    return nef_rdc_frame


def _read_file_or_exit_error(file_path):
    file_name = str(file_path)
    try:
        with file_path.open() as file_h:
            db_text = file_h.readlines()

    except Exception as e:
        exit_error(f"Error reading file {file_name} because {e}", e)
    return db_text


def _normalise_chain_codes(chain_codes, pales_input_files):
    if not pales_input_files:
        _exit_error_no_input_files()
    if not chain_codes:
        chain_codes = ["A"]
    if len(chain_codes) == 1 and len(pales_input_files) > 1:
        chain_codes = chain_codes * len(pales_input_files)
    elif len(chain_codes) != len(pales_input_files):
        _exit_error_number_chains_doesnt_match_number_files(
            chain_codes, pales_input_files
        )
    return chain_codes


def _exit_error_no_input_files():
    exit_error("No input files provided")


def _exit_error_number_chains_doesnt_match_number_files(chain_codes, pales_input_files):
    NEWLINE = "\n"
    input_file_names = [str(file) for file in pales_input_files]
    msg = f"""
            Number of chain codes must be zero or one or must be the same as the number of input files

            I got

            chain codes: {', '.join(chain_codes)}
            file_names:
            {NEWLINE.join(input_file_names)}
        """
    exit_error(msg)


def rdc_tensors_to_nef_loop(tensors: List[RdcTensorFrameData]) -> Loop:
    if isinstance(tensors, RdcTensorFrameData):
        tensors = [
            tensors,
        ]

    loop = Loop.from_scratch()
    loop.category = "_nefpls_tensor"
    loop.add_tag(
        [
            "index",
            "ex_x",
            "ex_y",
            "ex_z",
            "ey_x",
            "ey_y",
            "ey_z",
            "ez_x",
            "ez_y",
            "ez_z",
        ]
    )

    for index, tensor in enumerate(tensors, start=1):
        loop.add_data(
            [
                index,
                *tensor.eigen_vector_x,
                *tensor.eigen_vector_y,
                *tensor.eigen_vector_z,
            ]
        )
    return loop


def rdc_tensor_to_nef_saveframe(tensor: RdcTensorFrameData, frame_name) -> Saveframe:
    saveframe = create_nef_save_frame(_NEFL_RDC_TENSOR_FRAME_FRAME_CATEGORY, frame_name)

    saveframe.add_tag("Da", tensor.da)
    saveframe.add_tag("Dr", tensor.dr)

    loop = rdc_tensors_to_nef_loop([tensor])
    saveframe.add_loop(loop)

    return saveframe


def _parse_data_origin(db_records):
    mode = select_data_records(db_records, "PALES_MODE")

    result = UNUSED
    if mode:
        mode = mode[0].values[1]
        if mode in _PALES_SIMULATION_MODES:
            result = RestraintOrigin.PREDICTED
        elif mode == _PALES_ANALYSIS_MODES:
            result = RestraintOrigin.MEASURED
        else:
            UNUSED  # effectively i don't know
    return result


def _parse_rdc_tensor(db_records):

    raw_da = select_data_records(db_records, "Da")
    da = float(raw_da[0].values[1]) if raw_da else FLOAT_NAN

    raw_dr = select_data_records(db_records, "Dr")
    dr = float(raw_dr[0].values[1]) if raw_dr else FLOAT_NAN

    raw_eigen_vectors = select_data_records(db_records, "EIGENVECTORS")

    raw_eigen_vectors = {
        vector.values[1]: [float(elem) for elem in vector.values[2:]]
        for vector in raw_eigen_vectors
        if len(vector.values) > 1 and vector.values[1] in PALES_EIGENVECTOR_AXIS_LIST
    }

    eigen_vectors = [
        tuple(raw_eigen_vectors[key]) for key in sorted(raw_eigen_vectors.keys())
    ]
    if not eigen_vectors:
        eigen_vectors = [UNDEFINED_VECTOR_3D, UNDEFINED_VECTOR_3D, UNDEFINED_VECTOR_3D]

    return RdcTensorFrameData(da, dr, *eigen_vectors)


def _parse_rdcs(db_records, chain_codes, source):

    _exit_if_not_pales_rdc_file(db_records)

    chain_code_1, chain_code_2 = chain_codes

    rdcs = []

    data = select_records(db_records, VALUES)
    column_indices = get_column_indices(db_records)

    for index, line in enumerate(data, start=1):

        atom_name_1 = line.values[column_indices["ATOMNAME_I"]]
        residue_number_1 = line.values[column_indices["RESID_I"]]
        residue_type_1 = line.values[column_indices["RESNAME_I"]]

        atom_name_2 = line.values[column_indices["ATOMNAME_J"]]
        residue_number_2 = line.values[column_indices["RESID_J"]]
        residue_type_2 = line.values[column_indices["RESNAME_J"]]

        if source == PalesDataType.OBSERVED:
            rdc_value = line.values[column_indices["D_OBS"]]
            rdc_value_uncertainty = None
        else:
            rdc_value = line.values[column_indices["D"]]
            rdc_value_uncertainty = line.values[column_indices["DD"]]

        weight = line.values[column_indices["W"]]

        atom_1 = AtomLabel(
            SequenceResidue(chain_code_1, residue_number_1, residue_type_1), atom_name_1
        )

        atom_2 = AtomLabel(
            SequenceResidue(chain_code_2, residue_number_2, residue_type_2), atom_name_2
        )
        rdc = RdcRestraint(atom_1, atom_2, rdc_value, rdc_value_uncertainty, weight)

        rdcs.append(rdc)

    return rdcs


def _exit_if_not_pales_rdc_file(db_records):
    if not _check_is_pales_rdc_file(db_records):
        file_name = str(db_records.file_name)
        msg = f"""
                File {file_name} is not a valid PALES RDC file
                [which must have the columns {','.join(_PALES_RDCS_EXPECTED_FIELDS)}]
            """
        exit_error(msg)


def _check_is_pales_rdc_file(gdb_file):
    remarks = select_records(gdb_file, "REMARK")

    remark_strings = [remark.values for remark in remarks]
    return set(_PALES_EXPECTED_REMARKS).issubset(remark_strings)


# def _check_is_pales_rdc_file(gdb_file):
#     columns = set(get_gdb_columns(gdb_file))
#     expected_fields = set(_PALES_RDCS_EXPECTED_FIELDS)
#
#     return expected_fields.issubset(columns)


def _RDCS_to_nef_saveframe(rdcs: List[RdcRestraint], frame_name) -> Saveframe:
    saveframe = create_nef_save_frame(_NEF_RDC_RESTRAINT_FRAME_CATEGORY, frame_name)
    saveframe.add_tag("potential_type", UNUSED)

    if rdcs:
        loop = Loop.from_scratch()
        loop.category = "_" + _NEF_RDC_RESTRAINT_CATEGORY
        loop.add_tag(
            [
                "index",
                "restraint_id",
                "chain_code_1",
                "sequence_code_1",
                "residue_name_1",
                "atom_name_1",
                "chain_code_2",
                "sequence_code_2",
                "residue_name_2",
                "atom_name_2",
                "target_value",
                "target_value_uncertainty",
                "scale",
            ]
        )
        saveframe.add_loop(loop)

        for index, rdc in enumerate(rdcs, start=1):
            loop.add_data(
                [
                    index,
                    index,
                    rdc.atom_1.residue.chain_code,
                    rdc.atom_1.residue.sequence_code,
                    rdc.atom_1.residue.residue_name,
                    rdc.atom_1.atom_name,
                    rdc.atom_2.residue.chain_code,
                    rdc.atom_2.residue.sequence_code,
                    rdc.atom_2.residue.residue_name,
                    rdc.atom_2.atom_name,
                    rdc.value,
                    rdc.value_uncertainty,
                    rdc.weight,
                ]
            )

    return saveframe



import sys
from pathlib import Path
from typing import Iterable, List

# TODO: better support for assigned residues, improved handling of chains
import typer

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    TRANSLATIONS_3_1_PROTEIN,
    BadResidue,
    get_chain_code_iter,
    get_residue_name_from_lookup,
    translate_1_to_3,
)
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_float,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.mars import import_app

DEFAULT_PSEUDO_RESIDUE_PREFIX = "PR_"
SEPARATORS = "_-"

PSEUDO_RESIDUE = "PSEUDO_RESIDUE"

H = "H"
N = "N"
HA = "HA"
CA = "CA"
CB = "CB"
C = "CO"

EXPECTED_HEADINGS = {H, N, HA, CA, f"{CA}-1", CB, f"{CB}-1", C, f"{C}-1", f"{HA}-1"}

MARS_UNDEFINED = "-"
app = typer.Typer()

PSEUDO_RESIDUE_FORMAT = "<OPTIONAL-TEXT><NUMBER><OPTIONAL-TEXT>"

# TODO: maybe we want to add some of this to comments...
RECOGNISE_RESIDUES_HELP = f"""\
    if the second string in a pseudo residue name looks like contains a a 3 letter residue parse it and add it to the
    NEF shift pseudo residue names can be of the form  {PSEUDO_RESIDUE_FORMAT}, examples:
    include

    1
    PR_4?
    PR_5GLY
    PR_5GLY?

    in the third case if this flag is set the residue is set to GLY

    currently if multiple residue codes are detected they are all ignored (with a warning)
"""


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    chain_codes: str = typer.Option(
        "-",
        "--chain",
        help="chain codes [default -] can be a comma separated list or can be called mutiple times",
        metavar="<CHAIN-CODE>",
    ),
    entry_name: str = typer.Option(
        "mars", "-e", "--entry-name", help="a name for the entry"
    ),
    frame_name: str = typer.Option(
        "mars", "-f", "--frame-name", help="a name for the frame"
    ),
    prefix_to_strip: str = typer.Option(
        DEFAULT_PSEUDO_RESIDUE_PREFIX,
        "-s",
        "--strip-prefix",
        help="string to strip from the start of pseudo residues if present",
    ),
    parse_residues: bool = typer.Option(
        True,
        "--parse-residues",
        help=RECOGNISE_RESIDUES_HELP,
    ),
    treat_as_unassigned: bool = typer.Option(
        False,
        "--unassigned",
        help="treat all shifts as unassigned shifts",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type mars shifts.tab", metavar="<MARS-shifts>.tab"
    ),
):
    """- import shifts from MARS input shift file <mars-shifts>.tab"""

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name)

    chain_codes = parse_comma_separated_options(chain_codes)

    pipe(
        entry,
        chain_codes,
        frame_name,
        file_names,
        prefix_to_strip,
        parse_residues,
        treat_as_unassigned,
    )


def pipe(
    entry,
    chain_codes,
    frame_name,
    file_names,
    prefix_to_strip=DEFAULT_PSEUDO_RESIDUE_PREFIX,
    parse_residues=True,
    treat_as_unassigned=False,
):

    sparky_frames = []

    chain_code_iter = get_chain_code_iter(chain_codes)
    for file_name, chain_code in zip(file_names, chain_code_iter):

        with open(file_name) as lines:

            sparky_shifts = _parse_shifts(
                lines,
                chain_code=chain_code,
                file_name=file_name,
                prefix_to_strip=prefix_to_strip,
                parse_residues=parse_residues,
                treat_as_unassigned=treat_as_unassigned,
            )

            frame = shifts_to_nef_frame(sparky_shifts, frame_name)

            sparky_frames.append(frame)

    entry = add_frames_to_entry(entry, sparky_frames)

    print(entry)


def _convert_residue_type_to_3_let_or_exit(residue_type, line_info):
    try:
        residue_type = translate_1_to_3([residue_type])[0]
    except BadResidue:
        msg = f"""\
            unknown residue name {residue_type}
            at line {line_info.line_no} in
            file {line_info.file_name}, the line had the following content
            {line_info.line}
        """
        exit_error(msg)

    return residue_type


def _exit_if_residue_doesnt_equal_sequence(residue, chain_seqid_to_type, line_info):

    chain_code = residue.chain_code
    sequence_code = residue.sequence_code
    sequence_residue_name = get_residue_name_from_lookup(
        chain_code, sequence_code, chain_seqid_to_type
    )
    if sequence_residue_name.lower() != residue.residue_name.lower():
        msg = f"""\
            for residue number {sequence_code} in chain {chain_code} the residue type read from file
            {residue.residue_name} doesn't match the one from the sequence {sequence_residue_name}
            while reading line {line_info.line_no} from file:
            {line_info.file_name}
            the value of the line was
            {line_info.line}
        """
        exit_error(msg)


def _parse_shifts(
    lines: Iterable[str],
    chain_code: str = "-",
    file_name="unknown",
    prefix_to_strip=DEFAULT_PSEUDO_RESIDUE_PREFIX,
    parse_residues=True,
    treat_as_unassigned=False,
) -> ShiftList:

    heading_indices = {}

    read_shifts = []
    for i, line in enumerate(lines):
        line_info = LineInfo(file_name, i + 1, line)
        line = line.strip()

        if i == 0:
            headings = line.split()

            _exit_if_bad_header(line_info)

            heading_indices = heading_to_indices_dict(headings)

        elif not line:
            continue
        else:
            shifts = _parse_line(
                line_info,
                heading_indices,
                chain_code,
                prefix_to_strip,
                parse_residues,
                treat_as_unassigned,
            )

            read_shifts.extend(shifts)

    return ShiftList(read_shifts)


def heading_to_indices_dict(headings):
    heading_indices = {
        heading: column for column, heading in enumerate(headings, start=1)
    }

    heading_indices[PSEUDO_RESIDUE] = 0

    return heading_indices


def _get_residues_from_string(string):

    result = []
    string = string.upper()
    for residue_name in TRANSLATIONS_3_1_PROTEIN:
        if residue_name in string:
            result.append(residue_name)
    return result


# TODO: remove when python 3.8 no longer supported
def _removeprefix(target: str, prefix: str) -> str:
    if target.startswith(prefix):
        return target[len(prefix) :]
    else:
        return target


def _removesuffix(target: str, suffix: str) -> str:
    if target.endswith(suffix):
        return target[: -len(suffix)]
    else:
        return target


def _parse_line(
    line_info,
    heading_indices,
    chain_code,
    prefix_to_strip,
    parse_residues,
    treat_as_unassigned,
):
    results = []

    fields = line_info.line.strip().split()
    pseudo_residue = fields[heading_indices[PSEUDO_RESIDUE]]

    for heading, index in heading_indices.items():
        if heading != PSEUDO_RESIDUE:
            value = fields[index]
            if value != MARS_UNDEFINED:

                _exit_if_shift_is_not_float(value, line_info)

                value = float(value)

                pseudo_residue = _removeprefix(pseudo_residue, prefix_to_strip)

                (
                    pseudo_pre_chars,
                    pseudo_residue_number,
                    pseudo_post_chars,
                ) = _split_pseudo_residue(pseudo_residue)

                _exit_if_pseudo_residue_number_missing(
                    pseudo_residue_number, pseudo_residue, line_info
                )

                residue_names = []
                if parse_residues:
                    residue_names = _get_residues_from_string(pseudo_post_chars)
                    if len(residue_names) > 1:
                        _warn_multiple_residues(residue_names, line_info)

                residue_name = UNUSED if not residue_names else residue_names[0]
                atom_name, offset = _split_heading(heading)

                if pseudo_residue.startswith("AR_") and not treat_as_unassigned:
                    if offset == "":
                        residue = SequenceResidue(
                            chain_code=chain_code,
                            sequence_code=pseudo_residue_number,
                            residue_name=residue_name,
                        )
                    else:
                        continue
                else:
                    if treat_as_unassigned:
                        chain_code = "-"
                    residue = SequenceResidue(
                        chain_code=f"@{chain_code}",
                        sequence_code=f"@{pseudo_residue_number}{offset}",
                        residue_name=UNUSED,
                    )
                atom = AtomLabel(residue, atom_name=atom_name)

                shift = ShiftData(atom=atom, value=value)

                results.append(shift)

    return results


def _split_heading(heading: str):

    offset = ""
    if heading.endswith("-1"):
        heading = _removesuffix(heading, "-1")
        offset = "-1"
    return heading, offset


def _warn_multiple_residues(residue_names, line_info):
    msg = f"""\
                        WARNING: the pseudo residue on line {line_info.line_no} in file {line_info.file_name} contains
                                 more than one residue name and will be ignored, the residue names were:

                                 {' '.join(residue_names)}

                                 the complete line was:

                                 {line_info.line}
                    """
    print(msg, files=sys.stderr)


def _split_pseudo_residue(pseudo_residue):
    first_characters = _get_starting_alpha(pseudo_residue)

    first_characters = first_characters.rstrip(SEPARATORS)

    pseudo_residue = _removeprefix(pseudo_residue, first_characters)
    numbers = get_starting_number(pseudo_residue)

    last_characters = _removeprefix(pseudo_residue, numbers)
    last_characters = last_characters.lstrip(SEPARATORS)

    return first_characters, numbers, last_characters


def _exit_if_pseudo_residue_number_missing(
    pseudo_residue_number, pseudo_residue, line_info
):

    if not is_int(pseudo_residue_number):
        msg = f"""
            At line {line_info.line} in the file {line_info.file_name} the pseudo residue should include a number,
            I got {pseudo_residue_number} from the {pseudo_residue} which should have the format {PSEUDO_RESIDUE_FORMAT}
            the complete line was:

            {line_info.line}
        """
        exit_error(msg)


def _exit_if_shift_is_not_float(value, line_info):
    if not is_float(value):
        msg = f"""
                        at line {line_info.line_no} in file {line_info.file_name} i expected a chemical shift but the
                        value read {value} doesn't look like a floating point number, the complete line was
                        {line_info.line}
                    """

        exit_error(msg)


def _exit_if_wrong_number_data_columns(fields, headings, line_info):
    num_fields = len(fields)
    num_headings = len(headings)
    if num_fields < num_headings:
        msg = f"""\
                    at line {line_info.line_no} in
                    file {line_info.file_name}
                    there weren't enough columns of data i expected {num_headings} but got {num_fields}
                    so the number of fields matched the length of the headings.
                    the headings were {' '.join(headings)}
                    the line was {line_info.line}
                """
        exit_error(msg)


def _exit_if_bad_header(line_info):
    line = line_info.line
    headings = line_to_headings(line)

    if not set(headings).issubset(EXPECTED_HEADINGS):
        msg = f"""\
                    the file {line_info.file_name} doesn't look like a sparky file"
                    line 1 should contain the standard sparky heading:

                    {' '.join(EXPECTED_HEADINGS)}

                    i got:
                    {line_info.line}

                    note: the heading maybe a subset of the headings and typically should include at least H and N
                          and (typically) more than one of CA + CA-1, or CB + CB-1 or C + C-1
                """
        exit_error(msg)


def line_to_headings(line):
    return set(line.strip().split())


def _get_starting_alpha(group):
    result = []
    for char in group:
        if char.isalpha() or char in SEPARATORS:
            result.append(char)
        else:
            break
    return "".join(result)


def get_starting_number(group):
    result = []
    for char in group:
        if char.isdigit():
            result.append(char)
        else:
            break
    return "".join(result)
